﻿#pragma once
#include <memory>
#include <vector>
#include <algorithm>
#include "base_declaration.hpp"
#include "function_declaration.hpp"
#include "type.hpp"
#include "utilities.hpp"

namespace swer
{
    /**
        Represents declaration of the class in Swer, which is basicaly sequence of declarations with optional generics.
    */
    class class_declaration : public base_declaration
    {
    public:
        class_declaration(std::string&& identifier, base_declaration_seq&& declarations)
            : class_declaration(identifier_seq{}, std::move(identifier), std::move(declarations))
        {
        }

        class_declaration(identifier_seq&& types, std::string&& identifier, base_declaration_seq&& declarations)
            : m_identifiers(std::move(types)), m_identifier(std::move(identifier)), m_declarations(std::move(declarations))
        {
        }

        /**
            Returns name of the class.
        */
        const std::string& class_name() const
        {
            return m_identifier;
        }

        /**
            Returns names of generic arguments.
        */
        const identifier_seq& generic_identifiers() const
        {
            return m_identifiers;
        }

        std::string as_string() const override
        {
            class_declarations classes;
            classes.push_back(this);
            return concat(
                template_args_decl(m_identifiers),
                "struct ", m_identifier, "\n{\n",
                forward_declare_classes(),
                forward_declare_fncs(),
                declare_vars(),
                "};\n",
                define_classes(classes),
                define_fncs(classes)
            );
        }

        /**
            Generates a declaration of the class in C++.
        */
        std::string declare() const
        {
            return concat(
                template_args_decl(m_identifiers),
                "struct ", m_identifier, ";\n"
            );
        }

        /**
            Generates a definition of the class in C++.
            classes are parent classes of this class in terms of nesting.
        */
        std::string define(class_declarations& classes) const
        {
            std::string str;
            str += concat(
                string_join(classes.begin(), classes.end(), "", [](const class_declaration* c)
                {
                    return template_args_decl(c->generic_identifiers());
                }),
                template_args_decl(m_identifiers),
                "struct "
            );
            
            if (!classes.empty())
            {
                str += concat(string_join(classes.begin(), classes.end(), "", [](const class_declaration* c)
                    {
                        return concat(c->class_name(), template_args_use(c->generic_identifiers()), "::");
                    })
                );
            }

            classes.push_back(this);
            str += concat(m_identifier, "\n{\n",
                forward_declare_classes(),
                forward_declare_fncs(),
                declare_vars(),
                "};\n",
                define_classes(classes),
                define_fncs(classes)
            );
            classes.pop_back();
            return str;
        }

    private:
        /**
            Generates declarations of sub classes of this class.
        */
        std::string forward_declare_classes() const
        {
            std::string str;
            for_each_if(m_declarations.begin(), m_declarations.end(), is_class, [&str](const base_declaration_ptr& decl)
            {
                str += as_class(decl)->declare();
            });
            return str;
        }

        /**
            Generates declarations of sub functions of this class.
        */
        std::string forward_declare_fncs() const
        {
            std::string str;
            for_each_if(m_declarations.begin(), m_declarations.end(), &is_function, [&str](const base_declaration_ptr& decl)
            {
                str += as_function(decl)->declare();
            });
            return str;
        }

        /**
            Generates definitions of sub classes of this class.
        */
        std::string define_classes(class_declarations& classes) const
        {
            std::string str;
            for_each_if(m_declarations.begin(), m_declarations.end(), &is_class, [&](const base_declaration_ptr& decl)
            {
                str += as_class(decl)->define(classes);
            });
            return str;
        }

        /**
            Generates definitions of sub functions of this class.
        */
        std::string define_fncs(class_declarations& classes) const
        {
            std::string str;
            for_each_if(m_declarations.begin(), m_declarations.end(), &is_function, [&](const base_declaration_ptr& decl)
            {
                str += as_function(decl)->define(classes);
            });
            return str;
        }

        /**
            Generates declarations of variables in this class.
        */
        std::string declare_vars() const
        {
            std::string str;
            for_each_if(m_declarations.begin(), m_declarations.end(), &is_variable, [&str](const base_declaration_ptr& decl)
            {
                auto variable_decl = as_variable(decl);
                str += concat(decl->as_string(), ";\n");
            });
            return str;
        }

        std::vector<std::string> m_identifiers;
        std::string m_identifier;
        base_declaration_seq m_declarations;
    };
}
