﻿#pragma once
#include <string>
#include <fstream>
#include <memory>
#include "std_utils/make_unique.hpp"

namespace swer
{
    /**
        Supported C++ compilers.
        When adding new compiler, specialization of compiler::cpp_compilers with is also required.
    */
    enum class compilers
    {
        NOT_SELECTED,
        GCC, // Calls g++
        CLANG, // Calls clang++
        CL, // Calls cl.exe
    };

    /**
        Represents error occured when passing options to executable, such as mixing options, that can't be specified together.
    */
    struct options_error : std::runtime_error
    {
        explicit options_error(const std::string& msg) : std::runtime_error(msg)
        {
        }
    };

    /**
        Represents a class which holds all command line options.
    */
    class options
    {
    public:
        options() : m_compiler(compilers::NOT_SELECTED), m_command(""), m_output("")
        {
        }

        /**
            Returns value representing a call to g++
        */
        static std::string gcc_command()
        {
            return "g++ -std=c++14 -x c++ - ";
        }

        /**
        Returns value representing a call to clang++
        */
        static std::string clang_command()
        {
            return "clang++ -std=c++14 -x c++ - ";
        }

        /**
            Returns value representing a call to cl
        */
        static std::string cl_command()
        {
            return "cl /EHsc /Fea.exe";
        }

        /**
            Sets a command, -C option
        */
        void set_command(const std::string& command)
        {
            if (m_command != "")
                throw options_error("Option -C can't be specified more than once.");
            if (m_compiler != compilers::NOT_SELECTED)
                throw options_error("Option -C can't me mixed together with -c.");
            if (m_output != "")
                throw options_error("Option -C can't be mixed together with -o.");
            m_command = command;
        }
        
        /**
            Sets a compiler, -c option.
        */
        void set_compiler(compilers compiler)
        {
            if (m_compiler != compilers::NOT_SELECTED)
                throw options_error("Option -c can't be specified more than once.");
            if (m_command != "")
                throw options_error("Option -c can't me mixed together with -C.");
            m_compiler = compiler;
        }

        /**
            Sets an output file, -o option.
        */
        void set_output(const std::string& output)
        {
            if (m_output != "")
                throw options_error("Option -o can't be specified more than once.");
            if (m_command != "")
                throw options_error("Option -o can't be mixed together with -C");
            m_output = output;
        }

        /**
            Returns selected compiler.
        */
        compilers get_compiler() const
        {
            return m_compiler;
        }

        /**
            Returns provided command.
        */
        const std::string& get_command() const
        {
            return m_command;
        }

        /**
            Returns output file name.
        */
        const std::string& get_output() const
        {
            return m_output;
        }

        /**
            Returns either std::cout or ostream object with file prepared.
        */
        std::ostream& get_ostream(std::unique_ptr<std::ofstream>& str) const
        {
            if (m_output == "" || m_output == "-")
                return std::cout;
            str = std_utils::make_unique<std::ofstream>(m_output);
            return *str;
        }

    private:
        compilers m_compiler;
        std::string m_command;
        std::string m_output;
    };
}
