﻿#include <memory>
#include <ostream>
#include <string>
#include <functional>
#include "compiler.hpp"
#include "expression.hpp"
#include "error.hpp"

#ifndef _MSC_VER
#define L_tmpnam_s L_tmpnam
int tmpnam_s(char* buffer, size_t size)
{
    if (buffer == nullptr)
        return EINVAL;
    tmpnam(buffer);
    return 0;
}
#endif

std::string swer::compiler::get_iteration_identifier() const
{
    if (m_implicit_for_depth < MAX_IMPLICIT_FOR_DEPTH)
    {
        char c = static_cast<char>(FIRST_ITERATION_IDENTIFIER + m_implicit_for_depth);
        std::string identifier(1, c);
        if (is_variable_declared(identifier) != nullptr)
            throw compilation_error("Could not find any suitable identifier for iteration cycle with implicit control variable.");
        return identifier;
    }

    throw compilation_error(concat("You can not write more than ", std::to_string(MAX_IMPLICIT_FOR_DEPTH), " nested loops without defining variable explicitely."));
}

std::string swer::compiler::get_tmp_variable_identifier()
{
    return TMP_IDENTIFIER_BASE + std::to_string(m_tmp_identifier_order++);
}


void swer::compiler::declare_variables(const type_specifier_ptr& type, const identifier_initialization_seq& identifiers)
{
    for (auto& ident_init : identifiers)
    {
        if (!m_declared_identifiers.declare(m_stack_level, type, ident_init.identifier))
            global_logger->compilation_error(ident_init.location, concat("Variable ", ident_init.identifier, " was already defined."));
    }
}

const std::string swer::compiler::TMP_IDENTIFIER_BASE = "swer_tmp_ident_\\uAAAA_";


#ifdef _MSC_VER
inline FILE* popen(const char* command, const char* mode)
{
    return _popen(command, mode);
}
inline int pclose(FILE* file)
{
    return _pclose(file);
}
#endif

inline std::size_t file_write(FILE* file, const std::string& str)
{
    return fwrite(str.c_str(), sizeof(std::string::value_type), str.size(), file);
}

bool swer::compiler::generate_program_to_command(const std::string& command) const
{
    std::shared_ptr<FILE> cmd(popen(command.c_str(), "w"), pclose);
    if (!cmd)
        return false;

    return generate_program_by_function(std::bind(&file_write, cmd.get(), std::placeholders::_1));
}

bool swer::compiler::generate_program_to_stream(std::ostream& stream) const
{
    return generate_program_by_function(std::bind(&invoke_ostream_operator<std::string>, std::ref(stream), std::placeholders::_1));
}

// compiler::cpp_compiler specifiactions.
namespace swer
{
    template<>
    struct compiler::cpp_compiler<swer::compilers::GCC>
    {
        static bool generate_program(const swer::compiler& compiler, const std::string& output)
        {
            std::string command = "g++ -std=c++14 -x c++ - ";
            if (output != "")
                command += "-o " + output;
            return compiler.generate_program_to_command(command);
        }
    };

    template<>
    struct compiler::cpp_compiler<swer::compilers::CLANG>
    {
        static bool generate_program(const swer::compiler& compiler, const std::string& output)
        {
            std::string command = "clang++ -std=c++14 -x c++ - ";
            if (output != "")
                command += "-o " + output;
            return compiler.generate_program_to_command(command);
        }
    };

    template<>
    struct compiler::cpp_compiler<swer::compilers::CL>
    {
        static bool generate_program(const swer::compiler& compiler, const std::string& output)
        {
            // ToDo: file access rights problem on windows
            char file_name[L_tmpnam_s];
            tmpnam_s(file_name, L_tmpnam_s);
            std::ofstream file(file_name);
            if (!compiler.generate_program_to_stream(file))
                return false;

            std::string command = "cl /EHsc /TP " + std::string(file_name);
            if (output != "")
                command += " /Fe" + output;
            else
                command += " /Fea.exe";

            return std::system(command.c_str()) == 0;
        }
    };
}

bool swer::compiler::generate_program() const
{
    if (m_options.get_compiler() == compilers::NOT_SELECTED)
    {
        if (m_options.get_command() != "")
            return generate_program_to_command(m_options.get_command());
        std::unique_ptr<std::ofstream> str; // hidden delete
        return generate_program_to_stream(m_options.get_ostream(str));
    }

    switch (m_options.get_compiler())
    {
    case compilers::GCC:
        return cpp_compiler<compilers::GCC>::generate_program(*this, m_options.get_output());
    case compilers::CLANG:
        return cpp_compiler<compilers::CLANG>::generate_program(*this, m_options.get_output());
    case compilers::CL:
        return cpp_compiler<compilers::CL>::generate_program(*this, m_options.get_output());
    default:
        throw std::runtime_error("Unknown swer::compilers speicfication for cpp_compiler.");
    }
}
