﻿#pragma once
#include <memory>
#include <unordered_map>
#include <iostream>
#include <algorithm>
#include <set>
#include "type.hpp"
#include "base_expression.hpp"
#include "base_statement.hpp"
#include "base_declaration.hpp"
#include "class_declaration.hpp"
#include "variable_declaration.hpp"
#include "command_line_options.hpp"
#include "function_declaration.hpp"

namespace swer
{
    /**
        Represents a class, which controls generation of program.
        Also provides facilities for keeping track of current stack level and declared variables.

    */
    class compiler
    {
        /**
            Represents C++ compiler, which calls compiler::generate_program as required by C++ compiler.
        */
        template<compilers>
        struct cpp_compiler
        {
            static bool generate_program(const swer::compiler&, const std::string&)
            {
                static_assert(true, "compiler::cpp_compiler::generate_program was not specialized for such cpp_compiler.");
                return false;
            }
        };

    public:
        explicit compiler(const options& opt) : m_options(opt), m_stack_level(0), m_implicit_for_depth(0), m_tmp_identifier_order(0)
        {
        }

        /**
            Returns an identifier used in implicit for cycle statements. Returned variable depends on m_implicit_for_depth
            which is controlled by enter_implicit_for_block() and leave_implicit_for_block().
        */
        std::string get_iteration_identifier() const;

        /**
            Returns varibale with unique identifier with form if TMP_IDENTIFIER_BASE + m_tmp_identifier_order.
        */
        std::string get_tmp_variable_identifier();

        /**
            Declares a variable on current stack level.
        */
        void declare_variables(const type_specifier_ptr& type, const identifier_initialization_seq& identifiers);
        
        /**
            Checks whether variable is declared and returns it's type, or nullptr if note defined.
        */
        type_specifier_ptr is_variable_declared(const std::string& identifier) const
        {
            return m_declared_identifiers.is_declared(m_stack_level, identifier);
        }

        /**
            Enters new block.
        */
        void enter_block()
        {
            ++m_stack_level;
        }

        /**
            Un-declares variables at current block and leaves it.
        */
        void leave_block()
        {
            m_declared_identifiers.destroy(m_stack_level--);
        }

        /**
            Enters implicit for statement block.
        */
        void enter_implicit_for_block()
        {
            ++m_implicit_for_depth;
        }

        /**
            Leaves implicit for statement block.
        */
        void leave_implicit_for_block()
        {
            --m_implicit_for_depth;
        }

        /**
            Sets declarations of Swer program.
        */
        void set_declarations(base_declaration_seq&& declarations)
        {
            m_declarations = std::move(declarations);
        }

        /**
            Set statements of Swer program.
        */
        void set_statements(base_statement_seq&& statements)
        {
            m_statements = std::move(statements);
        }

        /**
            Generates program by values defined in m_options.
        */
        bool generate_program() const;

    private:
        /**
            Executes a command and writes a program to it's standard input.
        */
        bool generate_program_to_command(const std::string& command) const;

        /**
            Writes generated program to given stream.
        */
        bool generate_program_to_stream(std::ostream& stream) const;
        
        /**
            This functions controls the generation of Swer program to C++ source code, each part of the program is passed to generator function.
        */
        template<typename F>
        bool generate_program_by_function(F generator) const
        {
            generator(generate_includes());

            for_each_if(m_declarations.begin(), m_declarations.end(), &base_declaration::is_class, [&generator](const base_declaration_ptr& decl)
            {
                generator(base_declaration::as_class(decl)->declare());
            });

            for_each_if(m_declarations.begin(), m_declarations.end(), &base_declaration::is_function, [&generator](const base_declaration_ptr& decl)
            {
                generator(base_declaration::as_function(decl)->declare());
            });

            class_declarations c;
            for_each_if(m_declarations.begin(), m_declarations.end(), &base_declaration::is_class, [&generator,c](const base_declaration_ptr& decl) mutable
            {
                generator(base_declaration::as_class(decl)->define(c));
            });

            for_each_if(m_declarations.begin(), m_declarations.end(), &base_declaration::is_function, [&generator,c](const base_declaration_ptr& decl) mutable
            {
                generator(base_declaration::as_function(decl)->define(c));
            });


            generator(generate_main_intro());
            for (const auto& statement : m_statements)
                generator(statement->as_string());
            generator(generate_main_outro());
            return true;
        }

        /**
            Invokes ostream operator <<.
        */
        template<typename T>
        static auto invoke_ostream_operator(std::ostream& str, T&& t) -> decltype(str << std::forward<T>(t))
        {
            return str << std::forward<T>(t);
        }

        /**
            Generates intro of main function.
        */
        static std::string generate_main_intro()
        {
            return
R"(int main(int argc\uAAAA, char* argv\uAAAA[])
{
    Vector<String> _;
    for (int i\uAAAA = 1; i\uAAAA < argc\uAAAA; ++i\uAAAA)
        _.push_back(String(argv\uAAAA[i\uAAAA]));
    try
    {

)"
            ;
        }

        /**
            Generates outro of main function.
        */
        static std::string generate_main_outro()
        {
            return
R"(}
catch (const std::exception& e)
{
    print("Error: ", e.what());
}
}
)"
            ;
        }

        /**
            Generates required includes.
        */
        static std::string generate_includes()
        {
            return "#include <swer_stl>\n";
        }

        /**
            Custom comparer, which compares just identifiers and not types in declarations map.
        */
        using type_identifier = std::pair<type_specifier_ptr, std::string>;
        struct ident_set_comparer
        {
            bool operator()(const type_identifier& lhs, const type_identifier& rhs) const
            {
                return std::less<std::string>()(lhs.second, rhs.second);
            }
        };

        /**
            Represens a class, which keeps track of declared variables and provides facilities for declaring new one
            or removing all declared variables at given stack level.
        */
        class declaration_map
        {
        public:
            /**
                Declares new varibale on given stack level with type and name. Returns false if variable with such name was already defined.
            */
            bool declare(int stack_level, const type_specifier_ptr& type, const std::string& identifier)
            {
                return m_declared_variables[stack_level].emplace(type, identifier).second;
            }

            /**
                Removes all declarations of variables at given stack level.
            */
            void destroy(int stack_level)
            {
                m_declared_variables[stack_level].clear();
            }

            /**
                Checks, whether given variable is declared at given stack level or below (thus, is accessible), and returns its type,
                or nullptr if no such variable is declared.
            */
            type_specifier_ptr is_declared(int stack_level, const std::string& identifier) const
            {
                for (; stack_level >= 0; --stack_level)
                {
                    if (m_declared_variables.count(stack_level) == 0)
                        continue;
                    const auto& map = m_declared_variables.at(stack_level);
                    auto iter = std::find_if(map.cbegin(), map.cend(), [&identifier](decltype(*map.cbegin())& ident)
                    {
                        return ident.second == identifier;
                    });
                    if (iter != map.cend())
                        return iter->first;
                }
                return nullptr;
            }

        private:
            std::unordered_map<int, std::set<type_identifier, ident_set_comparer>> m_declared_variables;
        };

        const options& m_options;

        int m_stack_level;
        int m_implicit_for_depth;
        int m_tmp_identifier_order;
        declaration_map m_declared_identifiers;

        base_declaration_seq m_declarations;
        base_statement_seq m_statements;

        // Maximum for cycles with implicit variable declaration.
        constexpr static int MAX_IMPLICIT_FOR_DEPTH = 3;

        // First identifier for implicit iterations.
        constexpr static char FIRST_ITERATION_IDENTIFIER = 'i';
        const static std::string TMP_IDENTIFIER_BASE;
    };
}
