#pragma once
#include <string>
#include <unordered_map>

#include "command_line_options.hpp"

namespace swer
{ 
    /**
        Represents an error in compilation, which cannot be imediately reported.
    */
    struct compilation_error : std::runtime_error
    {
        explicit compilation_error(const std::string& msg) : std::runtime_error(msg)
        {
        }
    };

    /**
        Simple logger class allowing outputing to two streams (info and error) and keep track of number
        of encountered compilation errors and warnings.
    */
    class simple_logger
    {
    public:
        /**
            Constructs new logger with name of file being run, info stream and error stream
        */
        simple_logger(const std::string& exe_name, std::ostream& info_stream, std::ostream& error_stream)
            : m_exe_name(exe_name), m_info_stream(info_stream), m_error_stream(error_stream), m_warning(0), m_error(0)
        {
        }

        /**
            Repors an error encountered in the program.
        */
        void error(const std::string& message) const
        {
            m_error_stream << m_exe_name << ": error: " << message << std::endl;
        }

        /**
            Reports an error in compilation of the source file.
        */
        void compilation_error(unsigned line, const std::string& msg)
        {
            m_error_stream << "L" << line << ": " << msg << std::endl;
            ++m_error;
        }

        /**
            Reports a warning in compilation.
        */
        void compilation_warning(unsigned line, const std::string& msg)
        {
            m_error_stream << "L" << line << ": " << msg << std::endl;
            ++m_warning;
        }

        /**
            Returns count of encountered errors.
        */
        std::size_t compilation_error_count() const
        {
            return m_error;
        }

        /**
            Returns count of encountered warnings.
        */
        std::size_t compilation_warning_count() const
        {
            return m_warning;
        }

        /**
            Prints information about usage to info stream.
        */
        void info_usage() const
        {
            m_info_stream
                << "Usage: " << m_exe_name << " [options] file" << std::endl
                << "Attempts to compile file in Swer source code into C++ source code and writes it to stdout, unless -o, -c, -C is used." << std::endl
                << "Options: " << std::endl
                << " -h [--help]                    " << "Displays this help information." << std::endl
                << " -c [--compiler] [g++|clang|cl] " << "Redirects the output to compiler and invokes it." << std::endl
                << " -C [--command] <command>       " << "Redirects output to <command> and invokes it." << std::endl
                << " -o [--output] <file>           " << "Instead of writing to stdout, writes to file" << std::endl
                << std::endl
                << "-C and -c command can't be mixed. If -o is used with -c, then the -o option is passed to the specified compiler. -c is functionally equivalent to -C like this:" << std::endl
                << "-c g++   <=> -C \"" << options::gcc_command() << "\"" << std::endl
                << "-c clang <=> -C \"" << options::clang_command() << "\"" << std::endl
                << "-c cl    <=> -C \"" << options::cl_command() << "\"" << std::endl
                << std::endl
                << "For additional informations, see" << std::endl
                << "https://bitbucket.org/Zereges/swer" << std::endl
                ;
        }

    private:
        const std::string m_exe_name;
        std::ostream& m_info_stream;
        std::ostream& m_error_stream;

        std::size_t m_warning;
        std::size_t m_error;
    };

    extern ::std::unique_ptr<simple_logger> global_logger;

    inline void error(unsigned line, const std::string& msg)
    {
        global_logger->compilation_error(line, msg);
    }
}
