﻿#pragma once
#include <memory>
#include <vector>
#include <unordered_map>

#include "base_expression.hpp"
#include "variable_declaration.hpp"
#include "utilities.hpp"
#include "base_statement.hpp"
#include "numerics.hpp"
#include "operator_object.hpp"

namespace swer
{
    /**
        Represents an operator in Swer, that is used with expressions.

        Could be simplified by creating single class for each operator and overriding apply_operator, but it would make code longer
        and I would say less clear.
    */
    class operator_t
    {
    public:
        enum type
        {
            EXPRESSION,

            ADDITION,
            SUBTRACTION,
            MULTIPLICATION,
            DIVISION,
            MODULATION,
            EXPONENTIATION,
            UNARY_PLUS,
            UNARY_MINUS,
            PRE_INCREMENT,
            PRE_DECREMENT,
            POST_INCREMENT,
            POST_DECREMENT,
            FACTORIAL,

            ASSIGNMENT,
            ADDITION_ASSIGNMENT,
            SUBTRACTION_ASSIGNMENT,
            MULTIPLICATION_ASSIGNMENT,
            DIVISION_ASSIGNMENT,
            MODULATION_ASSIGNMENT,
            EXPONENTIATION_ASSIGNMENT,
            LEFT_SHIFT_ASSIGNMENT,
            RIGHT_SHIFT_ASSIGNMENT,
            BITWISE_AND_ASSIGNMENT,
            BITWISE_OR_ASSIGNMENT,
            BITWISE_XOR_ASSIGNMENT,

            LOGICAL_NOT,
            LOGICAL_AND,
            LOGICAL_OR,
            BITWISE_NOT,
            BITWISE_AND,
            BITWISE_OR,
            BITWISE_XOR,
            LEFT_SHIFT,
            RIGHT_SHIFT,

            CONDITIONAL,
            LESS,
            GREATER,
            LESS_EQUALS,
            GREATER_EQUALS,
            EQUALS,
            NOT_EQUALS,

            ADDRESS_OF,
            DEREFERENCE,
            MEMBER_ACCESS_DOT,
            MEMBER_ACCESS_RIGHT_ARROW,
            SUBSCRIPT,
            FUNCTION_CALL,
            EXPR_LIST,
            BRACED_LIST,
        };
#ifdef _MSC_VER
#pragma region operator_decls
#endif
        static std::shared_ptr<operator_t> expression;

        static std::shared_ptr<operator_t> addition;
        static std::shared_ptr<operator_t> subtraction;
        static std::shared_ptr<operator_t> multiplication;
        static std::shared_ptr<operator_t> division;
        static std::shared_ptr<operator_t> modulation;
        static std::shared_ptr<operator_t> exponentiation;
        static std::shared_ptr<operator_t> unary_plus;
        static std::shared_ptr<operator_t> unary_minus;
        static std::shared_ptr<operator_t> pre_increment;
        static std::shared_ptr<operator_t> pre_decrement;
        static std::shared_ptr<operator_t> post_increment;
        static std::shared_ptr<operator_t> post_decrement;
        static std::shared_ptr<operator_t> factorial;

        static std::shared_ptr<operator_t> assignment;
        static std::shared_ptr<operator_t> addition_assignment;
        static std::shared_ptr<operator_t> subtraction_assignment;
        static std::shared_ptr<operator_t> multiplication_assignment;
        static std::shared_ptr<operator_t> division_assignment;
        static std::shared_ptr<operator_t> modulation_assignment;
        static std::shared_ptr<operator_t> exponentiation_assignment;
        static std::shared_ptr<operator_t> left_shift_assignment;
        static std::shared_ptr<operator_t> right_shift_assignment;
        static std::shared_ptr<operator_t> bitwise_and_assignment;
        static std::shared_ptr<operator_t> bitwise_or_assignment;
        static std::shared_ptr<operator_t> bitwise_xor_assignment;

        static std::shared_ptr<operator_t> logical_not;
        static std::shared_ptr<operator_t> logical_and;
        static std::shared_ptr<operator_t> logical_or;
        static std::shared_ptr<operator_t> bitwise_not;
        static std::shared_ptr<operator_t> bitwise_and;
        static std::shared_ptr<operator_t> bitwise_or;
        static std::shared_ptr<operator_t> bitwise_xor;
        static std::shared_ptr<operator_t> left_shift;
        static std::shared_ptr<operator_t> right_shift;

        static std::shared_ptr<operator_t> conditional;
        static std::shared_ptr<operator_t> less;
        static std::shared_ptr<operator_t> greater;
        static std::shared_ptr<operator_t> less_equals;
        static std::shared_ptr<operator_t> greater_equals;
        static std::shared_ptr<operator_t> equals;
        static std::shared_ptr<operator_t> not_equals;

        static std::shared_ptr<operator_t> address_of;
        static std::shared_ptr<operator_t> dereference;
        static std::shared_ptr<operator_t> member_access_dot;
        static std::shared_ptr<operator_t> member_access_right_arrow;
        static std::shared_ptr<operator_t> subscript;
        static std::shared_ptr<operator_t> function_call;
        static std::shared_ptr<operator_t> expr_list;
        static std::shared_ptr<operator_t> braced_list;
#ifdef _MSC_VER
#pragma endregion operator_decls
#endif

        /**
            Constructs new operator based on type.
        */
        explicit operator_t(type t) : m_type(t)
        {
        }

        /**
            Aplies an operator to sequence of expressions.
        */
        std::string apply_operator(const base_expression_seq& exprs) const
        {
            switch (m_type)
            {
            case                type::EXPRESSION: return concat("(", exprs[0]->as_string(), ")");

            case                  type::ADDITION: return concat("(", exprs[0]->as_string(), " + ", exprs[1]->as_string(), ")");
            case               type::SUBTRACTION: return concat("(", exprs[0]->as_string(), " - ", exprs[1]->as_string(), ")");
            case            type::MULTIPLICATION: return concat("(", exprs[0]->as_string(), " * ", exprs[1]->as_string(), ")");
            case                  type::DIVISION: return concat("(", exprs[0]->as_string(), " / ", exprs[1]->as_string(), ")");
            case                type::MODULATION: return concat("(", exprs[0]->as_string(), " % ", exprs[1]->as_string(), ")");
            case            type::EXPONENTIATION: return concat("(swerstl::pow(", exprs[0]->as_string(), ", ", exprs[1]->as_string(), "))");
            case                type::UNARY_PLUS: return concat("(+", exprs[0]->as_string(), ")");
            case               type::UNARY_MINUS: return concat("(-", exprs[0]->as_string(), ")");
            case             type::PRE_INCREMENT: return concat("(++", exprs[0]->as_string(), ")");
            case             type::PRE_DECREMENT: return concat("(--", exprs[0]->as_string(), ")");
            case            type::POST_INCREMENT: return concat("(", exprs[0]->as_string(), "++)");
            case            type::POST_DECREMENT: return concat("(", exprs[0]->as_string(), "--)");
            case                 type::FACTORIAL: return concat("(swerstl::factorial(", exprs[0]->as_string(), "))");

            case                type::ASSIGNMENT: return concat("(", exprs[0]->as_string(), " = ", exprs[1]->as_string(), ")");
            case       type::ADDITION_ASSIGNMENT: return concat("(", exprs[0]->as_string(), " += ", exprs[1]->as_string(), ")");
            case    type::SUBTRACTION_ASSIGNMENT: return concat("(", exprs[0]->as_string(), " -= ", exprs[1]->as_string(), ")");
            case type::MULTIPLICATION_ASSIGNMENT: return concat("(", exprs[0]->as_string(), " *= ", exprs[1]->as_string(), ")");
            case       type::DIVISION_ASSIGNMENT: return concat("(", exprs[0]->as_string(), " /= ", exprs[1]->as_string(), ")");
            case     type::MODULATION_ASSIGNMENT: return concat("(", exprs[0]->as_string(), " %= ", exprs[1]->as_string(), ")");
            case type::EXPONENTIATION_ASSIGNMENT: return concat("(swerstl::pow_assign(", exprs[0]->as_string(), ", ", exprs[1]->as_string(), "))");
            case     type::LEFT_SHIFT_ASSIGNMENT: return concat("(", exprs[0]->as_string(), " <<= ", exprs[1]->as_string(), ")");
            case    type::RIGHT_SHIFT_ASSIGNMENT: return concat("(", exprs[0]->as_string(), " >>= ", exprs[1]->as_string(), ")");
            case    type::BITWISE_AND_ASSIGNMENT: return concat("(", exprs[0]->as_string(), " &= ", exprs[1]->as_string(), ")");
            case     type::BITWISE_OR_ASSIGNMENT: return concat("(", exprs[0]->as_string(), " |= ", exprs[1]->as_string(), ")");
            case    type::BITWISE_XOR_ASSIGNMENT: return concat("(", exprs[0]->as_string(), " ^= ", exprs[1]->as_string(), ")");

            case               type::LOGICAL_NOT: return concat("(!", exprs[0]->as_string(), ")");
            case               type::LOGICAL_AND: return concat("(", exprs[0]->as_string(), " && ", exprs[1]->as_string(), ")");
            case                type::LOGICAL_OR: return concat("(", exprs[0]->as_string(), " || ", exprs[1]->as_string(), ")");
            case               type::BITWISE_NOT: return concat("(~", exprs[0]->as_string(), ")");
            case               type::BITWISE_AND: return concat("(", exprs[0]->as_string(), " & ", exprs[1]->as_string(), ")");
            case                type::BITWISE_OR: return concat("(", exprs[0]->as_string(), " | ", exprs[1]->as_string(), ")");
            case               type::BITWISE_XOR: return concat("(", exprs[0]->as_string(), " ^ ", exprs[1]->as_string(), ")");
            case                type::LEFT_SHIFT: return concat("(", exprs[0]->as_string(), " << ", exprs[1]->as_string(), ")");
            case               type::RIGHT_SHIFT: return concat("(", exprs[0]->as_string(), " >> ", exprs[1]->as_string(), ")");

            case               type::CONDITIONAL: return concat("(", exprs[0]->as_string(), " ? ", exprs[1]->as_string(), " : ", exprs[2]->as_string(), ")");
            case                      type::LESS: return concat("(", exprs[0]->as_string(), " < ", exprs[1]->as_string(), ")");
            case                   type::GREATER: return concat("(", exprs[0]->as_string(), " > ", exprs[1]->as_string(), ")");
            case               type::LESS_EQUALS: return concat("(", exprs[0]->as_string(), " <= ", exprs[1]->as_string(), ")");
            case            type::GREATER_EQUALS: return concat("(", exprs[0]->as_string(), " >= ", exprs[1]->as_string(), ")");
            case                    type::EQUALS: return concat("(", exprs[0]->as_string(), " == ", exprs[1]->as_string(), ")");
            case                type::NOT_EQUALS: return concat("(", exprs[0]->as_string(), " != ", exprs[1]->as_string(), ")");

            case                type::ADDRESS_OF: return concat("(&", exprs[0]->as_string(), ")");
            case               type::DEREFERENCE: return concat("(*", exprs[0]->as_string(), ")");
            case         type::MEMBER_ACCESS_DOT: return concat(exprs[0]->as_string(), ".", exprs[1]->as_string());
            case type::MEMBER_ACCESS_RIGHT_ARROW: return concat(exprs[0]->as_string(), "->", exprs[1]->as_string());
            case                 type::SUBSCRIPT: return concat(exprs[0]->as_string(), "[", exprs[1]->as_string(), "]");
            case             type::FUNCTION_CALL: return concat(exprs[0]->as_string(), "(", exprs[1]->as_string(), ")");
            case                 type::EXPR_LIST: return string_join(exprs, ", ", as_string_decay<base_expression_ptr>);
            case               type::BRACED_LIST: return concat("{ ", string_join(exprs, ", ", as_string_decay<base_expression_ptr>), " }");
            }

            throw std::runtime_error("Can't happen. New operator addded, without apply_operator specififed.");
        }

    private:
        type m_type;
    };

    using operator_ptr = std::shared_ptr<operator_t>;

    /**
        Represents an expression in Swer. Expression has operator and sub-expressions.
    */
    class expression : public base_expression
    {
    public:
        template<typename... Expr>
        explicit expression(const operator_ptr& oper, Expr&&... exprs) : m_operator(oper)
        {
            populate_expr(std::forward<Expr>(exprs)...);
        }

        std::string as_string() const override
        {
            return m_operator->apply_operator(m_expressions);
        }

        /**
            Append additional expression to the end of the expression list.
        */
        void append_expression(base_expression_ptr&& expr)
        {
            m_expressions.push_back(std::move(expr));
        }

        const base_expression_ptr& first_expression() const
        {
            return m_expressions[0];
        }

    private:
        /**
            Populates vector of expressions by perfect forwarding to push_back.
        */
        template<typename FirstExpr, typename... Rest>
        void populate_expr(FirstExpr&& expr, Rest&&... rest)
        {
            m_expressions.push_back(std::forward<FirstExpr>(expr));
            populate_expr(std::forward<Rest>(rest)...);
        }

        /**
            End of recursion.
        */
        static void populate_expr()
        {
        }

        operator_ptr m_operator;
        base_expression_seq m_expressions;
    };

    using expression_ptr = std::shared_ptr<expression>;
}
