﻿#pragma once
#include <memory>
#include <vector>
#include <string>

#include "type.hpp"
#include "base_expression.hpp"
#include "base_statement.hpp"
#include "base_declaration.hpp"
#include "variable_declaration.hpp"
#include "utilities.hpp"

namespace swer
{
    class class_declaration;

    /**
        Represents a formal parameter as used in function. Formal parameter has type, name and optional defualt value.
    */
    class formal_parameter
    {
    public:
        explicit formal_parameter(type_specifier_ptr&& type)
            : formal_parameter(std::move(type), "", nullptr)
        {
            m_ident_init.with_equals = false;
        }

        formal_parameter(type_specifier_ptr&& type, std::string&& identifier)
            : formal_parameter(std::move(type), std::move(identifier), nullptr)
        {
            m_ident_init.with_equals = false;
        }

        formal_parameter(type_specifier_ptr&& type, std::string&& identifier, base_expression_ptr&& ident)
            : m_type(std::move(type))
        {
            m_ident_init.with_equals = true;
            m_ident_init.identifier = std::move(identifier);
            m_ident_init.initialization = std::move(ident);
        }

        /**
            Generates formal parameter in C++.
        */
        std::string as_string() const
        {
            return concat(m_type->as_string(), " ", m_ident_init.as_string());
        }

        /**
            Generates formal parameter in C++ while opting default value.
        */
        std::string as_string_no_init() const
        {
            return concat(m_type->as_string(), " ", m_ident_init.identifier);
        }
        

    private:
        type_specifier_ptr m_type;
        identifier_initialization m_ident_init;
    };

    using formal_parameter_ptr = std::shared_ptr<formal_parameter>;
    using formal_parameter_seq = std::vector<formal_parameter_ptr>;

    /**
        Represents a function in Swer, that has name, return value and sequence of formal parameters.
    */
    class function_declaration : public base_declaration
    {
    public:
        function_declaration(type_specifier_ptr&& type, std::string&& identifier, formal_parameter_seq&& params, base_statement_ptr&& body)
            : m_ret_type(std::move(type)), m_identifier(std::move(identifier)), m_formal_parameters(std::move(params)), m_body(std::move(body))
        {
        }

        std::string as_string() const override
        {
            return concat(
                m_ret_type->as_string(), " ", m_identifier, "(",
                string_join(m_formal_parameters, ", ", as_string_decay<formal_parameter_ptr>),
                ")\n", m_body->as_string(), "\n"
            );
        }

        /**
            Generates declaration of this function in C++.
        */
        std::string declare() const
        {
            return concat(m_ret_type->as_string(), " ", m_identifier, "(", string_join(m_formal_parameters, ", ", as_string_decay<formal_parameter_ptr>), ");\n");
        }

        /**
            Generates definition of this function in C++.
        */
        std::string define(class_declarations& classes) const;

    private:
        type_specifier_ptr m_ret_type;
        std::string m_identifier;
        formal_parameter_seq m_formal_parameters;
        base_statement_ptr m_body;
    };

    using function_ptr = std::shared_ptr<function_declaration>;
    using function_seq = std::vector<function_ptr>;
}
