﻿#include <iostream>
#include <fstream>
#include <getopt.h>
#include "error.hpp"
#include "grammar_defs.hpp"
#include "compiler.hpp"
#include "command_line_options.hpp"
#include "std_utils/make_unique.hpp"

namespace swer
{
    std::unique_ptr<simple_logger> global_logger;
}


#ifndef _MSC_VER
// just decay to potentionaly unsafe fopen
int fopen_s(FILE** ptr, const char* file, const char* mode)
{
    *ptr = fopen(file, mode);
    if (*ptr == nullptr)
        return EINVAL;
    return 0;
}
#endif

int main(int argc, char* argv[])
{
    swer::global_logger = std_utils::make_unique<swer::simple_logger>(argv[0], std::cerr, std::cerr);
    swer::options options;

    for (;;)
    {
        static option long_options[] =
        {
            {     "help",       no_argument, nullptr, 'h' },
            { "compiler", required_argument, nullptr, 'c' },
            {  "command", required_argument, nullptr, 'C' },
            {   "output", required_argument, nullptr, 'o' },
            {    nullptr,                 0, nullptr,  0  }
        };
        int option_index = 0;
        int c = getopt_long(argc, argv, "hC:c:o:", long_options, &option_index);
        if (c == -1)
            break;

        switch (c)
        {
        case 'h': // help
            swer::global_logger->info_usage();
            return swer::CODE_EXIT_SUCCESS;
        case 'c': // compiler specification
        {
            std::string compiler_str = optarg;
            swer::compilers compiler;
            if (compiler_str == "gcc" || compiler_str == "g++")
                compiler = swer::compilers::GCC;
            else if (compiler_str == "clang" || compiler_str == "clang++")
                compiler = swer::compilers::CLANG;
            else if (compiler_str == "cl")
                compiler = swer::compilers::CL;
            else
            {
                swer::global_logger->error("Invalid compiler specification for -c option. Expecting one of:\ng++ gcc\nclang++ clang\ncl");
                return swer::CODE_EXIT_FAILURE;
            }
            try
            {
                options.set_compiler(compiler);
            }
            catch (const swer::options_error& ex)
            {
                swer::global_logger->error(ex.what());
                return swer::CODE_EXIT_FAILURE;
            }
        }
        break;
        case 'C': // command specification
            try
            {
                options.set_command(optarg);
            }
            catch (const swer::options_error& ex)
            {
                swer::global_logger->error(ex.what());
                return swer::CODE_EXIT_FAILURE;
            }
            break;
        case 'o': // output
            try
            {
                options.set_output(optarg);
            }
            catch (const swer::options_error& ex)
            {
                swer::global_logger->error(ex.what());
                return swer::CODE_EXIT_FAILURE;
            }
            break;
        }
    }

    if (optind == argc)
    {
        swer::global_logger->error("No input file specified.");
        return swer::CODE_EXIT_FAILURE;
    }
    if (optind < argc - 1)
    {
        swer::global_logger->error("Too many arguments.");
        return swer::CODE_EXIT_FAILURE;
    }

    if (fopen_s(&yyin, argv[optind], "r"))
    {
        swer::global_logger->error(std::string(argv[optind]) + ": No such file.");
        return swer::CODE_EXIT_FAILURE;
    }

    swer::compiler comp(options);
    yyparse(&comp);
    fclose(yyin);

    std::size_t errs = swer::global_logger->compilation_error_count();

    try
    {
        if (!errs)
            return comp.generate_program() ? swer::CODE_EXIT_SUCCESS : swer::CODE_EXIT_FAILURE;
    }
    catch (const std::runtime_error& ex)
    {
        swer::global_logger->error(ex.what());
        return swer::CODE_EXIT_FAILURE;
    }
}
