%{
	#include <memory>
	#include <iostream>
	#include <string>
    #include <cstring>
	#include <queue>

	#include "../grammar_defs.hpp"
	#include "../error.hpp"
	#include "gen_bison.hpp"
	
	using namespace swer;
	unsigned line_no = 1;
	std::string cur_string;

	std::queue<int> token_queue;
%}

%option noyywrap
%option 8bit
%option	never-interactive
%option nounput

%s InChar
%s InString
%s InLineComment

WHITESPACE				[ \r\t\f]
IDENTIFIER				[a-zA-Z_][a-zA-Z_0-9]*

BINARY_LITERAL			0[bB][0-1]+
OCTAL_LITERAL			0[0-7]+
DECIMAL_LITERAL			0|[1-9][0-9]*
HEXADECIMAL_LITERAL		0[xX][0-9a-fA-F]+
FLOATING_LITERAL		([0-9]+\.)|([0-9]*\.?[0-9]+)([eE][+-]?[0-9]+)?

%%
	*line = line_no;
	if (!token_queue.empty())
	{
		auto token = std::move(token_queue.front());
		token_queue.pop();
		return token;
	}

<INITIAL>{
	\{							return LEFT_BRACE;
	\}							return RIGHT_BRACE;
	\[							return LEFT_BRACKET;
	\]							return RIGHT_BRACKET;
	\(							return LEFT_PARENTHESIS;
	\)							return RIGHT_PARENTHESIS;
	\<							return LEFT_CHEVRON;
	\>							return RIGHT_CHEVRON;
	
	~							return TILDE;
	::							return COLON_COLON;
	\.							return DOT;
	\.\.						return DOT_DOT;
	,							return COMMA;
	-\>							return RIGHT_ARROW;
	=\>							return RIGHT_RIGHT_ARROW;

	\+\+						return PLUS_PLUS;
	--							return MINUS_MINUS;

	\*							return ASTERISK;
	&							return AMPERSAND;
	\+							return PLUS;
	-							return MINUS;
	!							return NOT;
	#							return HASH;

	\/							return DIVIDE;
	%							return MODULO;

	\<\<						return LEFT_SHIFT;
	\>\>						return RIGHT_SHIFT;

	\<=							return LESS_EQUALS;
	\>=							return GREATER_EQUALS;

	==							return EQUALS_EQUALS;
	!=							return NOT_EQUALS;

	\^							return CARET;
	\^\^						return CARET_CARET;

	\|							return BAR;

	\?							return QUESTION_MARK;

	:							return COLON;
	;							return SEMICOLON;

	=							return EQUALS;
	\+=							return PLUS_EQUALS;
	-=							return MINUS_EQUALS;
	\*=							return ASTERISK_EQUALS;
	\/=							return DIVIDE_EQUALS;
	%=							return MODULO_EQUALS;
	\^=							return CARET_EQUALS;
	\<\<=						return LEFT_SHIFT_EQUALS;
	\>\>=						return RIGHT_SHIFT_EQUALS;
	&&=							return AMPERSAND_AMPERSAND_EQUALS;
	\|\|=						return BAR_BAR_EQUALS;
	\^\^=						return CARET_CARET_EQUALS;

	&&							return AMPERSAND_AMPERSAND;
	\|\|						return BAR_BAR;

	\$							return AUTO;
	
	@							return FOR;
	brk							return BREAK;
	cnt							return CONTINUE;
	if							return IF;
	el							return ELSE;
	sw							return SWITCH;
	

	'							BEGIN(InChar);
	\"							BEGIN(InString);
	\n							*line = ++line_no;

	{IDENTIFIER}				lval->string = yytext; return IDENTIFIER;

	\+\?						return OPERATOR_PLUS;
	-\?							return OPERATOR_MINUS;
	\*\?						return OPERATOR_MULTI;
	\/\?						return OPERATOR_DIVIDE;
	%\?							return OPERATOR_MODULO;
	\^\?						return OPERATOR_EXPONENT;

	{BINARY_LITERAL}			|
	{OCTAL_LITERAL}				|
	{DECIMAL_LITERAL}			|
	{HEXADECIMAL_LITERAL}		lval->number_i = std::stoi(yytext); return NUMBER_LITERAL;

	{FLOATING_LITERAL}			lval->number_f = std::stof(yytext); return FLOATING_LITERAL;
	
	/* 5..9 is parsed as 5. .9 */
	{DECIMAL_LITERAL}\.\.		token_queue.push(DOT_DOT); lval->number_i = std::stoi(std::string(yytext, yytext + std::strlen(yytext) - 2)); return NUMBER_LITERAL;
}

<InChar>{
	[^'\n]+						cur_string += yytext;
	\\'							cur_string += yytext;
	'							{
									BEGIN(INITIAL);
									if (cur_string.length() > 1)
										global_logger->compilation_error(line_no, "Too many characters in character literal.");
									if (cur_string.length() == 0)
									{
										global_logger->compilation_error(line_no, "Character literal should contain at least one character.");
										lval->string = " ";
									}
									else
										lval->string = std::string(1, cur_string[0]);
									cur_string = "";
									return CHARACTER_LITERAL;
								}
	\n							cur_string += "\n";
	<<EOF>>						{
									global_logger->compilation_error(line_no, "End of file in character literal.");
                                    *line = ++line_no;
									lval->string = std::move(cur_string);
									cur_string = "";
									BEGIN(INITIAL);
									return CHARACTER_LITERAL;
								}
}

<InString>{
	[^"\n]+						cur_string += yytext;
	\\\"						cur_string += R"(\")";
	\"							{
									lval->string = std::move(cur_string);
									cur_string = "";
									BEGIN(INITIAL);
									return STRING_LITERAL;
								}
	\n							cur_string += R"(\n)";
	<<EOF>>						{
									global_logger->compilation_error(line_no, "End of file in string literal.");
									lval->string = std::move(cur_string);
									cur_string = "";
									return STRING_LITERAL;
								}
}
<InLineComment>{
	\n							*line = ++line_no; BEGIN(INITIAL);
	.*	/* skip */
}

{WHITESPACE} /* skip */
\/\/							BEGIN(InLineComment);
