%{
	#include <utility>

	#include "../grammar_defs.hpp"
	#include "../error.hpp"

	#include "../literal.hpp"
	#include "../array.hpp"
	#include "../identifier.hpp"
	#include "../lambda.hpp"
	#include "../operator_object.hpp"
	
	#include "../case_statement.hpp"
	#include "../base_statement.hpp"
	#include "../plain_statement.hpp"
	#include "../switch_statement.hpp"
	#include "../for_statement.hpp"
	#include "../while_statement.hpp"
	#include "../for_range_statement.hpp"
	#include "../for_from_to_statement.hpp"
	#include "../iteration_statement.hpp"
	#include "../if_statement.hpp"
	#include "../keyword_statement.hpp"
	#include "../compound_statement.hpp"

	using namespace swer;

	inline void yyerror(unsigned* line, compiler* comp, const char* msg)
	{
		error(*line, msg);
	}


	#define YYERROR_VERBOSE 1

    /* locations */
    #define YYLTYPE unsigned
    #define YYLLOC_DEFAULT(Cur, Rhs, N)	\
		do								\
		{								\
			(Cur) = (Rhs)[0];			\
		} while(0)
%}

%token <string>		IDENTIFIER
%token <number_i>	NUMBER_LITERAL
%token <string>		CHARACTER_LITERAL
%token <string>		STRING_LITERAL
%token <number_f>	FLOATING_LITERAL

%token				LEFT_BRACE
%token				RIGHT_BRACE
%token				LEFT_BRACKET
%token				RIGHT_BRACKET
%token				LEFT_PARENTHESIS
%token				RIGHT_PARENTHESIS
%token				LEFT_CHEVRON
%token				RIGHT_CHEVRON

%token				TILDE
%token				COLON_COLON
%token				DOT
%token				DOT_DOT
%token				COMMA
%token				RIGHT_ARROW
%token				RIGHT_RIGHT_ARROW

%token				PLUS_PLUS
%token				MINUS_MINUS

%token				ASTERISK
%token				AMPERSAND
%token				PLUS
%token				MINUS
%token				NOT
%token				HASH

%token				DIVIDE
%token				MODULO

%token				OPERATOR_PLUS
%token				OPERATOR_MINUS
%token				OPERATOR_MULTI
%token				OPERATOR_DIVIDE
%token				OPERATOR_MODULO
%token				OPERATOR_EXPONENT

%token				LEFT_SHIFT
%token				RIGHT_SHIFT

%token				LESS_EQUALS
%token				GREATER_EQUALS

%token				EQUALS_EQUALS
%token				NOT_EQUALS

%token				CARET
%token				CARET_CARET

%token				BAR

%token				QUESTION_MARK

%token				COLON
%token				SEMICOLON

%token				EQUALS
%token				PLUS_EQUALS
%token				MINUS_EQUALS
%token				ASTERISK_EQUALS
%token				DIVIDE_EQUALS
%token				MODULO_EQUALS
%token				CARET_EQUALS
%token				LEFT_SHIFT_EQUALS
%token				RIGHT_SHIFT_EQUALS
%token				AMPERSAND_AMPERSAND_EQUALS
%token				CARET_CARET_EQUALS
%token				BAR_BAR_EQUALS

%token				AMPERSAND_AMPERSAND
%token				BAR_BAR

%token				AUTO

%token				SWITCH
%token				FOR
%token				BREAK
%token				CONTINUE
%token				RETURN
%token				IF
%token				ELSE


%type<type>					type_name type_specifier nested_name_specifier
%type<type_seq>				type_specifier_seq
%type<type_qualifier>		ref_or_ptr_qualifier
%type<type_qualifiers>		ref_or_ptr_qualifier_seq
%type<identifier_init>		identifier_with_opt_initialization initialization
%type<identifier_init_seq>	identifier_with_opt_initialization_seq
%type<declaration>			variable_declaration variable_declaration_with_initialization function_declaration class_declaration declaration
%type<declaration_seq>		class_body function_variable_class_declaration_seq declaration_seq
%type<formal_params>		formal_parameter
%type<formal_params_seq>	formal_parameter_list formal_parameters

%type<expression>			literal expression_base expression expression_postfix expression_assignment expression_logical_or expression_logical_and expression_bitwise_or
							expression_bitwise_xor expression_bitwise_and expression_comparsion expression_bitshift expression_addition expression_multiplication
							expression_exponentiation expression_prefix expression_constant initializer_clause initializer_list braced_init_list expression_list
							condition operator_object

%type<index>				operator_member_access operator_unary_addition operator_unary operator_exponentiation operator_multiplication operator_addition operator_shift
							operator_relational operator_assignment

%type<statement>			statement statement_dangling_else if_statement if_statement_dangling_else iteration_statement iteration_statement_dangling_else
							implicit_for_statement implicit_for_statement_dangling_else compound_statement declaration_statement switch_statement case_statement 
%type<statement_seq>		statement_seq case_statement_seq

%type<identifier_seq>		identifier_list lambda_introducer


%start swer
%error-verbose
%pure-parser
%parse-param {swer::compiler* comp}
%lex-param {swer::compiler* comp}
%locations

%%
swer
	: statement_seq
	{
		comp->set_statements(std::move($1));
	}
	| declaration_seq
	{
		comp->set_declarations(std::move($1));
	}
	| declaration_seq statement_seq
	{
		comp->set_declarations(std::move($1));
		comp->set_statements(std::move($2));
	}
	| /* empty */
	;

declaration_seq
	: declaration
	{
		$$.push_back(std::move($1));
	}
	| declaration_seq declaration
	{
		$$ = std::move($1);
		$$.push_back(std::move($2));
	}
	;

declaration
	: function_declaration
	{
		$$ = std::move($1);
	}
	| class_declaration
	{
		$$ = std::move($1);
	}
	;

function_declaration
	: type_specifier IDENTIFIER formal_parameters compound_statement
	{
		$$ = std::make_shared<function_declaration>(std::move($1), std::move($2), std::move($3), std::move($4));
	}
	;


type_specifier
	: type_name
	{
		$$ = std::move($1);
	}
	| type_name ref_or_ptr_qualifier_seq
	{
		$$ = std::move($1);
		for (auto&& qualifier : std::move($2))
			$$->append_qualifier(std::move(qualifier));
		$2.clear();
	}
	;

type_specifier_seq
	: type_specifier
	{
		$$.push_back(std::move($1));
	}
	| type_specifier_seq COMMA type_specifier
	{
		$$ = std::move($1);
		$$.push_back(std::move($3));
	}
	;

ref_or_ptr_qualifier_seq
	: ref_or_ptr_qualifier
	{
		$$.push_back(std::move($1));
	}
	| ref_or_ptr_qualifier_seq ref_or_ptr_qualifier
	{
		$$ = std::move($1);
		$$.push_back(std::move($2));
	}
	;

ref_or_ptr_qualifier
	: pointer
	{
		$$ = type::qualifier::pointer;
	}
	| reference
	{
		$$ = type::qualifier::reference;
	}
	;

pointer
	: HASH
	;

reference
	: TILDE
	;

/*
	todo auto not in typename (becasue nesting)
*/

type_name
	: AUTO
	{
		$$ = std::make_shared<type>(type::auto_type);
	}
	| IDENTIFIER
	{
		$$ = std::make_shared<type>(std::move($1));
	}
	| IDENTIFIER LEFT_BRACE type_specifier_seq RIGHT_BRACE
	{
		$$ = std::make_shared<type>(std::move($1), std::move($3));
	}
	| COLON_COLON IDENTIFIER
	{
		$$ = std::make_shared<type>(std::move($2));
		$$->prepend_qualifier(type::qualifier::nested);
	}
	| nested_name_specifier IDENTIFIER
	{
		$$ = std::make_shared<type>(std::move($1), std::move($2));
	}
	| COLON_COLON IDENTIFIER LEFT_BRACE type_specifier_seq RIGHT_BRACE
	{
		$$ = std::make_shared<type>(std::move($2), std::move($4));
		$$->prepend_qualifier(type::qualifier::nested);
	}
	| nested_name_specifier IDENTIFIER LEFT_BRACE type_specifier_seq RIGHT_BRACE
	{
		$$ = std::make_shared<type>(std::move($1), std::move($2), std::move($4));
	}
	;

nested_name_specifier
	: type_name COLON_COLON
	{
		$$ = std::move($1);
		$$->append_qualifier(type::qualifier::nested);
	}
	;



formal_parameters
	: LEFT_PARENTHESIS RIGHT_PARENTHESIS
	{
		$$ = formal_parameter_seq{};
	}
	| LEFT_PARENTHESIS formal_parameter_list RIGHT_PARENTHESIS
	{
		$$ = std::move($2);
	}
	;

formal_parameter_list
	: formal_parameter
	{
		$$.push_back(std::move($1));
	}
	| formal_parameter_list COMMA formal_parameter
	{
		$$ = std::move($1);
		$$.push_back(std::move($3));
	}
	;

formal_parameter
	: type_specifier
	{
		$$ = std::make_shared<formal_parameter>(std::move($1));
	}
	| type_specifier IDENTIFIER
	{
		$$ = std::make_shared<formal_parameter>(std::move($1), std::move($2));
	}
	| type_specifier IDENTIFIER EQUALS expression
	{
		$$ = std::make_shared<formal_parameter>(std::move($1), std::move($2), std::move($4));
	}
	| type_specifier IDENTIFIER EQUALS braced_init_list
	{
		$$ = std::make_shared<formal_parameter>(std::move($1), std::move($2), std::move($4));
	}
	;

braced_init_list
	: LEFT_BRACE RIGHT_BRACE
	{
		$$ = std::make_shared<expression>(operator_t::braced_list);
	}
	| LEFT_BRACE initializer_list RIGHT_BRACE
	{
		$$ = std::make_shared<expression>(operator_t::braced_list, std::move($2));
	}
	;

initializer_list
	: initializer_clause
	{
		$$ = std::make_shared<expression>(operator_t::expr_list, std::move($1));
	}
	| initializer_list COMMA initializer_clause
	{
		$$ = std::move($1);
		std::static_pointer_cast<expression>($$)->append_expression(std::move($3));
	}
	;

initializer_clause
	: expression
	{
		$$ = std::move($1);
	}
	| braced_init_list
	{
		$$ = std::move($1);
	}
	;


class_declaration
	: IDENTIFIER class_body
	{
		$$ = std::make_shared<class_declaration>(std::move($1), std::move($2));
	}
	| compound_statement_begin identifier_list compound_statement_end IDENTIFIER class_body
	{
		$$ = std::make_shared<class_declaration>(std::move($2), std::move($4), std::move($5));
	}
	;

identifier_list
	: IDENTIFIER
	{
		$$.push_back(std::move($1));
	}
	| identifier_list COMMA IDENTIFIER
	{
		$$ = std::move($1);
		$$.push_back(std::move($3));
	}

class_body
	: LEFT_BRACE RIGHT_BRACE
	{
		$$ = base_declaration_seq{};
	}
	| LEFT_BRACE function_variable_class_declaration_seq RIGHT_BRACE
	{
		$$ = std::move($2);
	}
	;

function_variable_class_declaration_seq
	: function_declaration
	{
		$$.push_back(std::move($1));
	}
	| variable_declaration SEMICOLON
	{
		$$.push_back(std::move($1));
	}
	| class_declaration
	{
		$$.push_back(std::move($1));
	}
	| function_variable_class_declaration_seq function_declaration
	{
		$$ = std::move($1);
		$$.push_back(std::move($2));
	}
	| function_variable_class_declaration_seq variable_declaration SEMICOLON
	{
		$$ = std::move($1);
		$$.push_back(std::move($2));
	}
	| function_variable_class_declaration_seq class_declaration
	{
		$$ = std::move($1);
		$$.push_back(std::move($2));
	}
	;


variable_declaration
	: type_specifier identifier_with_opt_initialization_seq
	{
		$$ = std::make_shared<variable_declaration>(std::move($1), std::move($2));
	}
	;

variable_declaration_with_initialization
	: type_specifier IDENTIFIER initialization
	{
		$3.identifier = std::move($2);
		$$ = std::make_shared<variable_declaration>(std::move($1), std::move($3));
	}
	;

identifier_with_opt_initialization_seq
	: identifier_with_opt_initialization
	{
		$$.push_back(std::move($1));
	}
	| identifier_with_opt_initialization_seq COMMA identifier_with_opt_initialization
	{
		$$ = std::move($1);
		$$.push_back(std::move($3));
	}
	;

identifier_with_opt_initialization
	: IDENTIFIER
	{
		$$.identifier = std::move($1);
		$$.location = @1;
	}
	| IDENTIFIER initialization
	{
		$$ = std::move($2);
		$$.identifier = std::move($1);
		$$.with_equals = $2.with_equals;
		$$.location = @1;
	}
	;

initialization
	: EQUALS expression
	{
		$$.with_equals = true;
		$$.initialization = std::move($2);
	}
	| braced_init_list
	{	
		$$.initialization = std::move($1);
	}
	;

/*
	ToDo: for cycle with multiple variables and no variables initialization
*/

statement
	: expression SEMICOLON
	{
		$$ = std::make_shared<plain_statement<base_expression>>(std::move($1));
	}
	| compound_statement
	{
		$$ = std::move($1);
	}
	| if_statement
	{
		$$ = std::move($1);
	}
	| switch_statement
	{
		$$ = std::move($1);
	}
	| iteration_statement
	{
		$$ = std::move($1);
	}
	| declaration_statement
	{
		$$ = std::move($1);
	}
	| BREAK SEMICOLON
	{
		$$ = std::make_shared<keyword_statement>(swer::keywords::BREAK);
	}
	| CONTINUE SEMICOLON
	{
		$$ = std::make_shared<keyword_statement>(swer::keywords::CONTINUE);
	}
	| return expression SEMICOLON
	{
		$$ = std::make_shared<keyword_statement>(swer::keywords::RETURN, std::move($2));
	}
	| return SEMICOLON
	{
		$$ = std::make_shared<keyword_statement>(swer::keywords::RETURN);
	}
	;

statement_dangling_else
	: expression SEMICOLON
	{
		$$ = std::make_shared<plain_statement<base_expression>>(std::move($1));
	}
	| compound_statement
	{
		$$ = std::move($1);
	}
	| switch_statement
	{
		$$ = std::move($1);
	}
	| if_statement_dangling_else
	{
		$$ = std::move($1);
	}
	| iteration_statement_dangling_else
	{
		$$ = std::move($1);
	}
	| declaration_statement
	{
		$$ = std::move($1);
	}
	| BREAK SEMICOLON
	{
		$$ = std::make_shared<keyword_statement>(swer::keywords::BREAK);
	}
	| CONTINUE SEMICOLON
	{
		$$ = std::make_shared<keyword_statement>(swer::keywords::CONTINUE);
	}
	| return expression SEMICOLON
	{
		$$ = std::make_shared<keyword_statement>(swer::keywords::RETURN, std::move($2));
	}
	| return SEMICOLON
	{
		$$ = std::make_shared<keyword_statement>(swer::keywords::RETURN);
	}
	;

return
	: RIGHT_ARROW
	;

compound_statement
	: LEFT_BRACE RIGHT_BRACE
	{
		$$ = std::make_shared<compound_statement>();
	}
	| compound_statement_begin statement_seq compound_statement_end
	{
		$$ = std::make_shared<compound_statement>(std::move($2));
	}
	;

compound_statement_begin
	: LEFT_BRACE
	{
		comp->enter_block();
	}
	;

compound_statement_end
	: RIGHT_BRACE
	{
		comp->leave_block();
	}
	;

statement_seq
	: statement
	{
		$$.push_back(std::move($1));
	}
	| statement_seq statement
	{
		$$ = std::move($1);
		$$.push_back(std::move($2));
	}
	;

if_statement
	: if_begin LEFT_PARENTHESIS condition RIGHT_PARENTHESIS statement if_end
	{
		$$ = std::make_shared<if_statement>(std::static_pointer_cast<condition>(std::move($3)), std::move($5));
	}
	| if_begin LEFT_PARENTHESIS condition RIGHT_PARENTHESIS statement_dangling_else ELSE statement if_end
	{
		$$ = std::make_shared<if_statement>(std::static_pointer_cast<condition>(std::move($3)), std::move($5), std::move($7));
	}
	;

if_statement_dangling_else
	: if_begin LEFT_PARENTHESIS condition RIGHT_PARENTHESIS statement_dangling_else ELSE statement_dangling_else if_end
	{
		$$ = std::make_shared<if_statement>(std::static_pointer_cast<condition>(std::move($3)), std::move($5), std::move($7));
	}
	;

if_begin
	: IF
	{
		comp->enter_block();
	}
	;

if_end
	: /* empty */
	{
		comp->leave_block();
	}
	;

switch_statement
	: SWITCH LEFT_PARENTHESIS condition RIGHT_PARENTHESIS LEFT_BRACE case_statement_seq RIGHT_BRACE
	{
		$$ = std::make_shared<switch_statement>(std::move($3), std::move($6));
	}
	;
	
case_statement_seq
	: case_statement
	{
		$$.push_back(std::move($1));
	}
	| case_statement_seq case_statement
	{
		$$ = std::move($1);
		$$.push_back(std::move($2));
	}
	;

case_statement
	: COLON expression_constant COLON statement_seq
	{
		$$ = std::make_shared<case_statement>(std::move($2), std::move($4));
	}
	| COLON ELSE COLON statement_seq
	{
		$$ = std::make_shared<case_statement>(std::move($4));
	}
	| COLON expression_constant COLON
	{
		$$ = std::make_shared<case_statement>(std::move($2));
	}
	| COLON ELSE COLON
	{
		$$ = std::make_shared<case_statement>();
	}
	;

iteration_statement
	: implicit_for_statement
	{
		$$ = std::move($1);
	}
	| for_begin IDENTIFIER LEFT_PARENTHESIS expression_constant DOT_DOT expression_constant RIGHT_PARENTHESIS statement for_end
	{
		$$ = std::make_shared<for_from_to_statement>(std::move($2), std::move($4), std::move($6), std::move($8), comp);
	}
	| for_begin IDENTIFIER LEFT_PARENTHESIS expression RIGHT_PARENTHESIS statement for_end
	{
		$$ = std::make_shared<for_range_statement>(std::move($2), std::move($4), std::move($6));
	}
	| for_begin LEFT_PARENTHESIS expression SEMICOLON RIGHT_PARENTHESIS statement for_end
	{
		$$ = std::make_shared<while_statement>(std::move($3), std::move($6));
	}
	| for_begin LEFT_PARENTHESIS expression SEMICOLON expression SEMICOLON expression_list RIGHT_PARENTHESIS statement for_end
	{
		$$ = std::make_shared<for_statement>(std::move($3), std::move($5), std::move($7), std::move($9));
	}
	| for_begin LEFT_PARENTHESIS variable_declaration_with_initialization SEMICOLON expression SEMICOLON expression_list RIGHT_PARENTHESIS statement for_end
	{
		$$ = std::make_shared<for_statement>(std::static_pointer_cast<variable_declaration>(std::move($3)), std::move($5), std::move($7), std::move($9));
	}
	;

implicit_for_statement
	: for_begin LEFT_PARENTHESIS expression_constant DOT_DOT expression_constant RIGHT_PARENTHESIS implicit_for_begin statement implicit_for_end for_end
	{
        try
		{
			$$ = std::make_shared<for_from_to_statement>(std::move($3), std::move($5), std::move($8), comp);
		}
		catch (const compilation_error& e)
		{
			$$ = nullptr;
			global_logger->compilation_error(@1, e.what());
		}
	}
	| for_begin LEFT_PARENTHESIS expression RIGHT_PARENTHESIS implicit_for_begin statement implicit_for_end for_end
	{
        try
		{
			$$ = std::make_shared<for_range_statement>(std::move($3), std::move($6), comp);
		}
		catch (const compilation_error& e)
		{
			$$ = nullptr;
			global_logger->compilation_error(@1, e.what());
		}
	}
	;

iteration_statement_dangling_else
	: implicit_for_statement_dangling_else
	{
		$$ = std::move($1);
	}
	| for_begin IDENTIFIER LEFT_PARENTHESIS expression_constant DOT_DOT expression_constant RIGHT_PARENTHESIS statement_dangling_else for_end
	{
		$$ = std::make_shared<for_from_to_statement>(std::move($2), std::move($4), std::move($6), std::move($8), comp);
	}
	| for_begin IDENTIFIER LEFT_PARENTHESIS expression RIGHT_PARENTHESIS statement_dangling_else for_end
	{
		$$ = std::make_shared<for_range_statement>(std::move($2), std::move($4), std::move($6));
	}
	| for_begin LEFT_PARENTHESIS expression SEMICOLON RIGHT_PARENTHESIS statement_dangling_else for_end
	{
		$$ = std::make_shared<while_statement>(std::move($3), std::move($6));
	}
	| for_begin LEFT_PARENTHESIS expression SEMICOLON expression SEMICOLON expression_list RIGHT_PARENTHESIS statement_dangling_else for_end
	{
		$$ = std::make_shared<for_statement>(std::move($3), std::move($5), std::move($7), std::move($9));
	}
	| for_begin LEFT_PARENTHESIS variable_declaration_with_initialization SEMICOLON expression SEMICOLON expression_list RIGHT_PARENTHESIS statement_dangling_else for_end
	{
		$$ = std::make_shared<for_statement>(std::static_pointer_cast<variable_declaration>(std::move($3)), std::move($5), std::move($7), std::move($9));
	}
	;

implicit_for_statement_dangling_else
	: for_begin LEFT_PARENTHESIS expression_constant DOT_DOT expression_constant RIGHT_PARENTHESIS implicit_for_begin statement_dangling_else implicit_for_end for_end
	{
		try
		{
			$$ = std::make_shared<for_from_to_statement>(std::move($3), std::move($5), std::move($8), comp);
		}
		catch (const compilation_error& e)
		{
			$$ = nullptr;
			global_logger->compilation_error(@1, e.what());
		}
	}
	| for_begin LEFT_PARENTHESIS expression RIGHT_PARENTHESIS implicit_for_begin statement_dangling_else implicit_for_end for_end
	{
		try
		{
			$$ = std::make_shared<for_range_statement>(std::move($3), std::move($6), comp);
		}
		catch (const compilation_error& e)
		{
			$$ = nullptr;
			global_logger->compilation_error(@1, e.what());
		}
	}
	;


implicit_for_begin
	: /* empty */
	{
		comp->enter_implicit_for_block();	
	}
	;


for_begin
	: FOR 
	{
		@$ = @1;
		comp->enter_block();
	}
	;

for_end
	: /* empty */
	{
		comp->leave_block();
	}
	;

implicit_for_end
	: /* empty */
	{
		comp->leave_implicit_for_block();
	}
	;

declaration_statement
	: variable_declaration SEMICOLON
	{
		auto variable = std::static_pointer_cast<variable_declaration>(std::move($1));
		comp->declare_variables(variable->get_type(), variable->get_identifier_initializations());
		$$ = std::make_shared<plain_statement<variable_declaration>>(std::move(variable));
	}
	;


condition
	: expression
	{
		$$ = std::make_shared<condition>(std::move($1));
	}
	| variable_declaration_with_initialization
	{
		$$ = std::make_shared<condition>(std::static_pointer_cast<variable_declaration>(std::move($1)));
	}
	;

expression_list
	: expression
	{
		$$ = std::make_shared<expression>(operator_t::expr_list, std::move($1));
	}
	| expression_list COMMA expression
	{
		$$ = std::move($1);
		std::static_pointer_cast<expression>($$)->append_expression(std::move($3));
	}
	;


expression
	: expression_assignment
	{
		$$ = std::move($1);
	}
	;

expression_assignment
	: expression_logical_or
	{
		$$ = std::move($1);
	}
	| expression_logical_or QUESTION_MARK expression COLON expression_assignment
	{
		$$ = std::make_shared<expression>(operator_t::conditional, std::move($1), std::move($3), std::move($5));
	}
	| expression_logical_or operator_assignment initializer_clause
	{
		std::shared_ptr<operator_t> oper;
		switch ($2)
		{
		case EQUALS: oper = operator_t::assignment; break;
		case MODULO_EQUALS: oper = operator_t::modulation_assignment; break;
		case PLUS_EQUALS: oper = operator_t::addition_assignment; break;
		case ASTERISK_EQUALS: oper = operator_t::multiplication_assignment; break;
		case DIVIDE_EQUALS: oper = operator_t::division_assignment; break;
		case MINUS_EQUALS: oper = operator_t::subtraction_assignment; break;
		case CARET_EQUALS: oper = operator_t::exponentiation_assignment; break;
		case LEFT_SHIFT_EQUALS: oper = operator_t::left_shift_assignment; break;
		case RIGHT_SHIFT_EQUALS: oper = operator_t::right_shift_assignment; break;
		case AMPERSAND_AMPERSAND_EQUALS: oper = operator_t::bitwise_and_assignment; break;
		case BAR_BAR_EQUALS: oper = operator_t::bitwise_or_assignment; break;
		case CARET_CARET_EQUALS: oper = operator_t::bitwise_xor_assignment; break;
		default: throw std::runtime_error("Invalid operator type for unary.");
		}
		$$ = std::make_shared<expression>(std::move(oper), std::move($1), std::move($3));
	}
	;

operator_assignment
	: EQUALS { $$ = EQUALS; }
	| PLUS_EQUALS { $$ = PLUS_EQUALS; }
	| MINUS_EQUALS { $$ = MINUS_EQUALS; }
	| ASTERISK_EQUALS { $$ = ASTERISK_EQUALS; }
	| DIVIDE_EQUALS { $$ = DIVIDE_EQUALS; }
	| MODULO_EQUALS { $$ = MODULO_EQUALS; }
	| CARET_EQUALS { $$ = CARET_EQUALS; }
	| LEFT_SHIFT_EQUALS { $$ = LEFT_SHIFT_EQUALS; }
	| RIGHT_SHIFT_EQUALS { $$ = RIGHT_SHIFT_EQUALS; }
	| AMPERSAND_AMPERSAND_EQUALS { $$ = AMPERSAND_AMPERSAND_EQUALS; }
	| BAR_BAR_EQUALS { $$ = BAR_BAR_EQUALS; }
	| CARET_CARET_EQUALS { $$ = CARET_CARET_EQUALS; }
	;

expression_logical_or
	: expression_logical_and
	{
		$$ = std::move($1);
	}
	| expression_logical_or BAR expression_logical_and
	{
		$$ = std::make_shared<expression>(operator_t::logical_or, std::move($1), std::move($3));
	}
	;

expression_logical_and
	: expression_bitwise_or
	{
		$$ = std::move($1);
	}
	| expression_logical_and AMPERSAND expression_bitwise_or
	{
		$$ = std::make_shared<expression>(operator_t::logical_and, std::move($1), std::move($3));
	}
	;

expression_bitwise_or
	: expression_bitwise_xor
	{
		$$ = std::move($1);
	}
	| expression_bitwise_or BAR_BAR expression_bitwise_xor
	{
		$$ = std::make_shared<expression>(operator_t::bitwise_or, std::move($1), std::move($3));
	}
	;

expression_bitwise_xor
	: expression_bitwise_and
	{
		$$ = std::move($1);
	}
	| expression_bitwise_xor CARET_CARET expression_bitwise_and
	{
		$$ = std::make_shared<expression>(operator_t::bitwise_xor, std::move($1), std::move($3));
	}
	;

expression_bitwise_and
	: expression_comparsion
	{
		$$ = std::move($1);
	}
	| expression_bitwise_and AMPERSAND_AMPERSAND expression_comparsion
	{
		$$ = std::make_shared<expression>(operator_t::bitwise_and, std::move($1), std::move($3));
	}
	;

expression_comparsion
	: expression_bitshift
	{
		$$ = std::move($1);
	}
	| expression_comparsion operator_relational expression_bitshift
	{
		std::shared_ptr<operator_t> oper;
		switch ($2)
		{
		case LEFT_CHEVRON: oper = operator_t::less; break;
		case RIGHT_CHEVRON: oper = operator_t::greater; break;
		case LESS_EQUALS: oper = operator_t::less_equals; break;
		case GREATER_EQUALS: oper = operator_t::greater_equals; break;
		case EQUALS_EQUALS: oper = operator_t::equals; break;
		case NOT_EQUALS: oper = operator_t::not_equals; break;
		default: throw std::runtime_error("Invalid operator type for unary.");
		}
		$$ = std::make_shared<expression>(std::move(oper), std::move($1), std::move($3));
	}
	;

operator_relational
	: LEFT_CHEVRON { $$ = LEFT_CHEVRON; }
	| RIGHT_CHEVRON { $$ = RIGHT_CHEVRON; }
	| LESS_EQUALS { $$ = LESS_EQUALS; }
	| GREATER_EQUALS { $$ = GREATER_EQUALS; }
	| EQUALS_EQUALS { $$ = EQUALS_EQUALS; }
	| NOT_EQUALS { $$ = NOT_EQUALS; }
	;

expression_bitshift
	: expression_addition
	{
		$$ = std::move($1);
	}
	| expression_bitshift operator_shift expression_addition
	{
		if ($2 == LEFT_SHIFT)
			$$ = std::make_shared<expression>(operator_t::left_shift, std::move($1), std::move($3));		
		else if ($2 == RIGHT_SHIFT)
			$$ = std::make_shared<expression>(operator_t::right_shift, std::move($1), std::move($3));		
		else
			throw std::runtime_error("Invalid operator type for bitshift.");
	}
	;

operator_shift
	: LEFT_SHIFT { $$ = LEFT_SHIFT; }
	| RIGHT_SHIFT { $$ = RIGHT_SHIFT; }
	;

expression_addition
	: expression_multiplication
	{
		$$ = std::move($1);
	}
	| expression_addition operator_addition expression_multiplication
	{
		if ($2 == PLUS)
			$$ = std::make_shared<expression>(operator_t::addition, std::move($1), std::move($3));
		else if ($2 == MINUS)
			$$ = std::make_shared<expression>(operator_t::subtraction, std::move($1), std::move($3));
		else
			throw std::runtime_error("Invalid operator type for addition.");
	}
	;

operator_addition
	: PLUS { $$ = PLUS; }
	| MINUS { $$ = MINUS; }
	;

expression_multiplication
	: expression_exponentiation
	{
		$$ = std::move($1);
	}
	| expression_multiplication operator_multiplication expression_exponentiation
	{
		if ($2 == ASTERISK)
			$$ = std::make_shared<expression>(operator_t::multiplication, std::move($1), std::move($3));
		else if ($2 == DIVIDE)
			$$ = std::make_shared<expression>(operator_t::division, std::move($1), std::move($3));
		else if ($2 == MODULO)
			$$ = std::make_shared<expression>(operator_t::modulation, std::move($1), std::move($3));
		else
			throw std::runtime_error("Invalid operator type for multiplication.");
	}
	;

operator_multiplication
	: ASTERISK { $$ = ASTERISK; }
	| DIVIDE { $$ = DIVIDE; }
	| MODULO { $$ = MODULO; }
	;

expression_exponentiation
	: expression_prefix
	{
		$$ = std::move($1);
	}
	| expression_exponentiation operator_exponentiation expression_prefix
	{
		$$ = std::make_shared<expression>(operator_t::exponentiation, std::move($1), std::move($3));
	}
	;

operator_exponentiation
	: CARET { $$ = CARET; }
	;

expression_prefix
	: expression_postfix
	{
		$$ = std::move($1);
	}
	| operator_unary expression_prefix
	{
		std::shared_ptr<operator_t> oper;
		switch ($1)
		{
		case PLUS: oper = operator_t::unary_plus; break;
		case MINUS: oper = operator_t::unary_minus; break;
		case PLUS_PLUS: oper = operator_t::pre_increment; break;
		case MINUS_MINUS: oper = operator_t::pre_decrement; break;
		case NOT: oper = operator_t::logical_not; break;
		case TILDE: oper = operator_t::bitwise_not; break;
		case ASTERISK: oper = operator_t::dereference; break;
		case AMPERSAND: oper = operator_t::address_of; break;
		default: throw std::runtime_error("Invalid operator type for unary.");
		}
		$$ = std::make_shared<expression>(std::move(oper), std::move($2));
	}
	;


operator_unary
	: PLUS { $$ = PLUS; }
	| MINUS { $$ = MINUS; }
	| operator_unary_addition { $$ = $1; }
	| NOT { $$ = NOT; }
	| TILDE { $$ = TILDE; }
	| ASTERISK { $$ = ASTERISK; }
	| AMPERSAND	 { $$ = AMPERSAND; }
	;

expression_postfix
	: expression_base
	{
		$$ = std::move($1);
	}
	| expression_postfix LEFT_PARENTHESIS RIGHT_PARENTHESIS
	{
		$$ = std::make_shared<expression>(operator_t::function_call, std::move($1), std::make_shared<expression>(operator_t::expr_list));
	}
	| expression_postfix LEFT_PARENTHESIS initializer_list RIGHT_PARENTHESIS
	{
		$$ = std::make_shared<expression>(operator_t::function_call, std::move($1), std::move($3));
	}
	| expression_postfix LEFT_BRACKET expression RIGHT_BRACKET
	{
		$$ = std::make_shared<expression>(operator_t::subscript, std::move($1), std::move($3));
	}
	| expression_postfix operator_member_access IDENTIFIER
	{
		if ($2 == DOT)
			$$ = std::make_shared<expression>(operator_t::member_access_dot, std::move($1), std::make_shared<identifier>(std::move($3)));
		else if ($2 == RIGHT_ARROW)
			$$ = std::make_shared<expression>(operator_t::member_access_right_arrow, std::move($1), std::make_shared<identifier>(std::move($3)));
		else
			throw std::runtime_error("Invalid operator type for member access.");
	}
	| expression_postfix operator_unary_addition
	{
		if ($2 == PLUS_PLUS)
			$$ = std::make_shared<expression>(operator_t::post_increment, std::move($1));
		else if ($2 == MINUS_MINUS)
			$$ = std::make_shared<expression>(operator_t::post_decrement, std::move($1));
		else
			throw std::runtime_error("Invalid operator type for unary addition.");
	}
	| expression_postfix NOT
	{
		$$ = std::make_shared<expression>(operator_t::factorial, std::move($1));
	}
	;

operator_unary_addition
	: PLUS_PLUS { $$ = PLUS_PLUS; }
	| MINUS_MINUS { $$ = MINUS_MINUS; }
	;

operator_member_access
	: DOT { $$ = DOT; }
	| RIGHT_ARROW { $$ = RIGHT_ARROW; }
	;

expression_base
	: literal
	{
		$$ = std::move($1);
	}
	| IDENTIFIER
	{
		$$ = std::make_shared<identifier>(std::move($1));
	}
	| LEFT_PARENTHESIS expression RIGHT_PARENTHESIS
	{
		$$ = std::make_shared<expression>(operator_t::expression, std::move($2));
	}
	| lambda_introducer RIGHT_RIGHT_ARROW compound_statement
	{
		$$ = std::make_shared<lambda>(std::move($1), std::move($3));
	}
	| operator_object
	{
		$$ = std::move($1);
	}
	;

operator_object
	: OPERATOR_PLUS
	{
		$$ = std::make_shared<operator_object>(operator_object::operator_type::PLUS);
	}
	| OPERATOR_MINUS
	{
		$$ = std::make_shared<operator_object>(operator_object::operator_type::MINUS);
	}
	| OPERATOR_MULTI
	{
		$$ = std::make_shared<operator_object>(operator_object::operator_type::MULTI);
	}
	| OPERATOR_DIVIDE
	{
		$$ = std::make_shared<operator_object>(operator_object::operator_type::DIVIDE);
	}
	| OPERATOR_MODULO
	{
		$$ = std::make_shared<operator_object>(operator_object::operator_type::MODULO);
	}
	| OPERATOR_EXPONENT
	{
		$$ = std::make_shared<operator_object>(operator_object::operator_type::EXPONENT);
	}
	;

lambda_introducer
	: LEFT_PARENTHESIS RIGHT_PARENTHESIS
	{
	}
	| IDENTIFIER
	{
		$$.push_back(std::move($1));
	}
	| LEFT_PARENTHESIS IDENTIFIER COMMA identifier_list RIGHT_PARENTHESIS
	{
		$$.push_back(std::move($2));
		for (auto iter = $4.begin(); iter != $4.end(); ++iter)
			$$.push_back(std::move(*iter));
	}

literal
	: NUMBER_LITERAL
	{
		$$ = std::make_shared<literal<integer>>(std::move($1));
	}
	| CHARACTER_LITERAL
	{
		char c = std::move($1)[0];
		$$ = std::make_shared<literal<char>>(std::move(c));
	}
	| FLOATING_LITERAL
	{
		$$ = std::make_shared<literal<floating>>(std::move($1));
	}
	| STRING_LITERAL
	{
		$$ = std::make_shared<literal<std::string>>(std::move($1));
	}
	| LEFT_BRACKET expression_list RIGHT_BRACKET
	{
		$$ = std::make_shared<array>(std::move($2));
	}
	;

expression_constant
	: expression_logical_or
	{
		$$ = std::move($1);
	}
	;
