﻿#pragma once
#include <string>
#include <deque>
#include <memory>
#include <vector>
#include <functional>
#include "utilities.hpp"

namespace swer
{
    /**
        Represents a type in Swer. Supports qualifiers, nesting and even generics.
    */
    class type
    {
    public:
        /**
            Represents a qualifier of the type.
        */
        class qualifier
        {
            /**
                All types of the qualifiers. Swer does not differentiete between qualifier used as prefixes or suffixes.
            */
            enum class qualifier_types
            {
                POINTER, // type#
                REFERENCE, // type~
                NESTED, // type:: or ::type
            };

        public:
            using this_ptr = std::shared_ptr<qualifier>;
            using this_seq = std::vector<this_ptr>;

            static this_ptr pointer;
            static this_ptr reference;
            static this_ptr nested;

            explicit qualifier(qualifier_types t) : m_qualifier(t)
            {
            }

            friend bool operator==(const qualifier& lhs, const qualifier& rhs)
            {
                return lhs.m_qualifier == rhs.m_qualifier;
            }

            std::string as_string() const
            {
                switch (m_qualifier)
                {
                case qualifier_types::POINTER: return "*";
                case qualifier_types::REFERENCE: return "&";
                case qualifier_types::NESTED: return "::";
                }
                throw std::runtime_error("Can't happen. New qualifier addded, without string representation.");
            }

        private:
            qualifier_types m_qualifier;
        };

        /**
            Type specifier, used in generic arguments.
        */
        using specifier_ptr = std::shared_ptr<type>;
        
        /**
            Type qualifier.
        */
        using qualifier_ptr = qualifier::this_ptr;
        using specifier_seq = std::vector<specifier_ptr>;
        using qualifier_seq = qualifier::this_seq;

        explicit type(std::string&& idenfitier)
            : m_identifier(std::move(idenfitier))
        {
        }

        explicit type(const std::string& idenfitier)
            : m_identifier(idenfitier)
        {
        }

        type(std::string&& idenfitier, specifier_seq&& generics)
            : m_identifier(std::move(idenfitier)), m_generics(std::move(generics))
        {
        }

        type(specifier_ptr&& nested_type, std::string&& idenfitier)
            : m_nested_type(std::move(nested_type)), m_identifier(std::move(idenfitier))
        {
        }

        type(specifier_ptr&& nested_type, std::string&& idenfitier, specifier_seq&& generics)
            : m_nested_type(std::move(nested_type)), m_identifier(std::move(idenfitier)), m_generics(std::move(generics))
        {
        }

        /**
            Textual representation of the type as to be generated in C++.
        */
        std::string as_string() const
        {
            std::string str;
            str += string_join(m_prefixes, "", as_string_decay<qualifier_ptr>);
            if (m_nested_type != nullptr)
                str += m_nested_type->as_string();
            str += m_identifier;

            if (m_generics.size() > 0)
                str += concat("<", string_join(m_generics, ", ", as_string_decay<specifier_ptr>), ">");

            str += string_join(m_suffixes, "", as_string_decay<qualifier_ptr>);
            return std::move(str);
        }

        /**
            Appends qualifier at the end of the type.
        */
        void append_qualifier(const qualifier_ptr& qualifier)
        {
            m_suffixes.push_back(std::move(qualifier));
        }

        /**
            Prepends qualifier on the beginning of the type.
        */
        void prepend_qualifier(const qualifier_ptr& qualifier)
        {
            m_prefixes.push_front(std::move(qualifier));
        }

        const static std::string auto_type;

    private:
        std::deque<qualifier_ptr> m_prefixes;
        std::deque<qualifier_ptr> m_suffixes;

        specifier_ptr m_nested_type;
        std::string m_identifier;
        specifier_seq m_generics;
    };
    using type_specifier_ptr = type::specifier_ptr;
    using type_qualifier_ptr = type::qualifier_ptr;
    using type_specifier_seq = type::specifier_seq;
    using type_qualifier_seq = type::qualifier_seq;
}
