﻿#pragma once
#include <vector>
#include <string>

namespace swer
{
    using identifier_seq = std::vector<std::string>;

    /**
        Concatenates all arguments into one std::string.
    */
    template<typename... String>
    std::string concat(const std::string& str, const String&... strs)
    {
        return str + concat(strs...);
    }

    inline std::string concat()
    {
        return "";
    }

    /**
        Joins elements defined by sequence of iterators together, separated by delimiter and each element is
        converted to std::string using to_string_fnc.
    */
    template<typename Iterator, typename Fnc>
    std::string string_join(Iterator begin, Iterator end, const std::string& delimiter, Fnc to_string_fnc)
    {
        std::string str;
        for (auto iter = begin; iter != end; ++iter)
            str += concat(to_string_fnc(*iter), delimiter);
        if (std::distance(begin, end) > 0)
            str.erase(str.size() - delimiter.size(), delimiter.size());
        return str;
    }

    /**
        Supposes elements are string and need no conversion.
    */
    template<typename Iterator>
    std::string string_join(Iterator begin, Iterator end, const std::string& delimiter)
    {
        return string_join(begin, end, delimiter, [](const std::string& item) { return item; });
    }

    template<typename Container, typename Fnc>
    std::string string_join(const Container& container, const std::string& delimiter, Fnc to_string_fnc)
    {
        return string_join(container.begin(), container.end(), delimiter, to_string_fnc);
    }

    template<typename Container>
    std::string string_join(const Container& container, const std::string& delimiter)
    {
        return string_join(container.begin(), container.end(), delimiter);
    }


    namespace detail
    {
        template<typename T>
        struct is_shared_ptr : std::false_type
        {
        };

        template<typename T>
        struct is_shared_ptr<std::shared_ptr<T>> : std::true_type
        {
        };

        template<typename Ptr>
        std::string as_string_decay(std::true_type, const std::shared_ptr<Ptr>& ptr)
        {
            return ptr->as_string();
        }

        template<typename T>
        std::string as_string_decay(std::false_type, const T& t)
        {
            return t.as_string();
        }
    }

    /**
        Calls t->as_string() (if decltype(t) is shared_ptr or t.as_string() (otherwise).
    */
    template<typename T>
    std::string as_string_decay(const T& t)
    {
        return detail::as_string_decay(detail::is_shared_ptr<T>{}, t);
    }

    /**
        For each cycle, which calls function only on elements, which satisfies a condition.
    */
    template<typename Iterator, typename Cond, typename Fnc>
    void for_each_if(Iterator begin, Iterator end, Cond condition, Fnc function)
    {
        for (Iterator iter = begin; iter != end; ++iter)
        {
            if (condition(*iter))
                function(*iter);
        }
    }

    /**
        Generates template declarations by arguments. (template<typename A, typename B, typename C>)
    */
    inline std::string template_args_decl(const std::vector<std::string>& generic_idents)
    {
        if (!generic_idents.empty())
            return concat("template<typename ", string_join(generic_idents, ", typename "), ">\n");
        return "";
    }

    /**
        Generates template usage by arguments (<A, B, C>)
    */
    inline std::string template_args_use(const std::vector<std::string>& generic_idents)
    {
        if (!generic_idents.empty())
            return concat("<", string_join(generic_idents, ","), ">");
        return "";
    }
}
