﻿#pragma once
#include <string>
#include "type.hpp"
#include "base_expression.hpp"
#include "base_declaration.hpp"
#include "utilities.hpp"

namespace swer
{
    /**
        Represents a name, with oprtional initialization. Tracks if initialization was done by = expr; or {expr};
    */
    struct identifier_initialization
    {
        identifier_initialization()
            : with_equals(false), initialization(nullptr), location(0)
        {
        }

        std::string as_string() const
        {
            std::string str;
            str += identifier;
            if (with_equals)
                str += " = ";
            if (initialization != nullptr)
                str += initialization->as_string();
            return std::move(str);
        }

        bool with_equals;
        std::string identifier;
        base_expression_ptr initialization;
        unsigned location;
    };

    using identifier_initialization_seq = std::vector<identifier_initialization>;
    
    /**
        Represents variable declaration in Swer, that is type and sequence of identifiers with optional initializations.
    */
    class variable_declaration : public base_declaration
    {
    public:
        variable_declaration(type_specifier_ptr&& type, identifier_initialization&& ident_init)
            : m_type(std::move(type)), m_identifiers({ std::move(ident_init) })
        {
        }

        variable_declaration(type_specifier_ptr&& type, identifier_initialization_seq&& seq)
            : m_type(std::move(type)), m_identifiers(std::move(seq))
        {
        }

        /**
            Generates declaration of variables in C++.
        */
        std::string as_string() const override
        {
            return concat(
                m_type->as_string(),
                " ",
                string_join(m_identifiers, ", ", [](const identifier_initialization& ident_init)
                {
                    std::string str = ident_init.as_string();
                    if (ident_init.initialization == nullptr)
                        str += "{}";
                    return str;
                })
            );
        }

        /**
            Returns type of this variables.
        */
        const type_specifier_ptr& get_type() const
        {
            return m_type;
        }

        /**
            Returns sequence of identifiers with initializatons.
        */
        const identifier_initialization_seq& get_identifier_initializations() const
        {
            return m_identifiers;
        }

    private:
        type_specifier_ptr m_type;
        identifier_initialization_seq m_identifiers;
    };
    using variable_declaration_ptr = std::shared_ptr<variable_declaration>;
}
