﻿#pragma once
#include "vector.hpp"
#include "numerics.hpp"
#include <functional>
#include "string.hpp"

namespace swerstl
{
    /**
        Generates a Vector, which containing [from, from+1, ..., to-1, to]
    */
    template<typename T>
    Vector<T> generate(const T& from, const T& to)
    {
        Vector<T> result;
        for (T i = from; i <= to; ++i)
            result.push_back(i);
        return result;
    }
    template<typename T>
    Vector<T> generate(const T& to)
    {
        return generate(1, to);
    }

    template<typename T>
    Vector<T> gen(const T& from, const T& to)
    {
        return generate(from, to);
    }

    template<typename T>
    Vector<T> gen(const T& to)
    {
        return generate(to);
    }

    /**
        Zip function as seen in functional programming. Combines two containers into one container, for which [i] = fnc(c1, c2);
        Has size of min(c1.size(), c2.size())
    */
    template<typename Container, typename F>
    Container zip(const Container& c1, const Container& c2, F fnc)
    {
        Container res;
        for (auto iter1 = c1.begin(), iter2 = c2.begin(); iter1 < c1.end() && iter2 < c2.end(); ++iter1, ++iter2)
            res.push_back(fnc(*iter1, *iter2));
        return res;
    }

    /**
        Map function as seen in functional programming. Constructs new container, for which [i] = fnc(cont[i]).
    */
    template<typename Container, typename F>
    Container map(const Container& cont, F fnc)
    {
        Container res = cont;
        for (auto& item : res)
            fnc(cont);
        return res;
    }

    /**
        Returns new container, which contains elements starting at cont[start_pos], for which cond(element) is true.
        Once cond(elements) become false, new container is returned.
    */
    template<typename Container, typename F>
    Container take_while(const Container& cont, Integer start_pos, F cond)
    {
        Container res;
        for (std::size_t i = start_pos; i < cont.size(); ++i)
            if (cond(cont[i]))
                res.push_back(cont[i]);
            else
                break;
        return res;
    }

    template<typename Container, typename F>
    Container take_while(const Container& vec, F cond)
    {
        return take_while(vec, 0, cond);
    }

    template<typename Container, typename F>
    Container tkw(const Container& vec, Integer start_pos, F cond)
    {
        return take_while(vec, start_pos, cond);
    }

    template<typename Container, typename F>
    Container tkw(const Container& vec, F cond)
    {
        return take_while(vec, cond);
    }

    /**
        Similar function to take_while, but all elements of cont, for which cond(element) is true are returned in new container of same type.
    */
    template<typename Container, typename F>
    Container filter(const Container& cont, F cond)
    {
        Container result;
        for (typename Container::const_reference item : cont)
            if (cond(item))
                result.push_back(item);
        return result;
    }

    template<typename Container, typename F>
    Container flt(const Container& vec, F fnc)
    {
        return filter(vec, fnc);
    }

    /**
        Fold function as seen in functional programming. Returns value, which is computed as
        fnc(fnc(...(fnc(cont[0], cont[1]), cont[2], ..., cont[cont.size()-1])...)).
        If cont is empty, returns default value for element type.
    */
    template<typename Container, typename F>
    typename Container::value_type fold(const Container& cont, F fnc)
    {
        if (cont.empty())
            return typename Container::value_type{};
        typename Container::value_type t = cont[0];
        for (auto iter = cont.begin() + 1; iter != cont.end(); ++iter)
            t = fnc(t, *iter);
        return t;
    }

    template<typename Container, typename F>
    typename Container::value_type fld(const Container& vec, F fnc)
    {
        return fold(vec, fnc);
    }

    /**
        Returns base to the power of exp.
    */
    inline Integer pow(Integer base, Integer exp)
    {
        Integer result = 1;
        for (int i = 0; i < exp; ++i)
            result *= base;
        return result;
    }

    inline Floating pow(Floating base, Floating exp)
    {
        return std::pow(base, exp);
    }

    inline Integer& pow_assign(Integer& number, Integer exp)
    {
        return number = pow(number, exp);
    }

    inline Floating& pow_assign(Floating& number, Floating exp)
    {
        return number = pow(number, exp);
    }


    /**
        Returns factorial of value.
    */
    inline Integer factorial(Integer value)
    {
        Integer result = 1;
        for (Integer i = 2; i <= value; ++i)
            result *= i;
        return result;
    }
}
