﻿#pragma once
#include <unordered_map>
#include <memory>
#include "pair.hpp"
#include "vector.hpp"

namespace swerstl
{
    namespace detail
    {
        /**
            Returns std::unique_ptr constructed by perfect forwarding args to constructor of T.
            std::make_unique not part of C++11.

            Only works for non array T's (missing SFINAE overload for T[])
        */
        template<typename T, typename... Args>
        std::unique_ptr<T> make_unique(Args&&... args)
        {
            return std::unique_ptr<T>(new T(std::forward<Args>(args)...));
        }
    }

    /**
        Map container stores pairs of Key, Mapped, where Key is used to access element Mapped.
    */
    template<typename Key, typename Mapped>
    class Map
    {
    public:
        using underlaying_container = std::unordered_map<Key, Mapped>;
        using key_type = Key;
        using mapped_type = Mapped;
        using value_type = Pair<Key, Mapped>;
        using reference = mapped_type&;
        using pointer = value_type*;
        using size_type = typename underlaying_container::size_type;
        using difference_type = typename underlaying_container::difference_type;

        /**
            Wrapper around std::unordered_map::iterator, which returns swerstl::pair when derefrenced.
            Does not fully satisfy Iterator requirements of C++ Standard.
        */
        struct iterator
        {
        private:
            using underlaying_iterator = typename underlaying_container::iterator;
        public:
            using difference_type = std::ptrdiff_t;
            using value_type = Pair<const Key&, Mapped&>;
            using reference = typename Map<Key, Mapped>::reference;
            using pointer = typename Map<Key, Mapped>::pointer;

            explicit iterator(underlaying_iterator&& iter) : m_iterator(std::move(iter))
            {
            }

            explicit iterator(const underlaying_iterator& iter) : m_iterator(iter)
            {
            }

            iterator operator++(int)
            {
                return iterator(m_iterator++);
            }
            iterator& operator++()
            {
                ++m_iterator;
                return *this;
            }

            iterator operator--(int)
            {
                return iterator(m_iterator--);
            }
            iterator& operator--()
            {
                --m_iterator;
                return *this;
            }

            value_type& operator*()
            {
                m_temp_pair = detail::make_unique<value_type>(m_iterator->first, m_iterator->second);
                return *m_temp_pair;
            }

            friend bool operator==(const iterator& lhs, const iterator& rhs) { return lhs.m_iterator == rhs.m_iterator; }
            friend bool operator!=(const iterator& lhs, const iterator& rhs) { return lhs.m_iterator != rhs.m_iterator; }
            
            pointer operator->() { return &operator*(); }

        private:
            underlaying_iterator m_iterator;
            std::unique_ptr<value_type> m_temp_pair;
        };

        Map() : m_map()
        {
            update_getters();
        }

        Map(const Map& other) : m_map(other.m_map)
        {
            update_getters();
        }

        Map(Map&& other) : m_map(std::move(other.m_map))
        {
            update_getters();
        }

        Map& operator=(const Map& other)
        {
            if (this != &other)
            {
                m_map = other.m_map;
                update_getters();
            }
            return *this;
        }

        Map& operator=(Map&& other)
        {
            if (this != &other)
            {
                m_map = std::move(other.m_map);
                update_getters();
            }
            return *this;
        }

        ~Map() = default;

        // ToDo const iters
        iterator begin() { return iterator(m_map.begin()); }
        iterator end() { return iterator(m_map.end()); }

        bool empty() const { return m_map.empty(); }
        size_type size() const { return m_map.size(); }
        size_type max_size() const { return m_map.max_size(); }

        void clear() { return m_map.clear(); }

        iterator erase(iterator pos) { return iterator(m_map.erase(pos.m_iterator)); }
        size_type erase(const key_type& key) { return m_map.erase(key); }

        reference operator[](const key_type& key)
        {
            return m_map[key];
        }

        /**
            Copies all mapped values to Vector and returns it.
            No specified order is guaranteed.
        */
        Vector<mapped_type> values()
        {
            Vector<mapped_type> res;
            for (auto val : *this)
                res.push_back(val.s);
            return res;
        }
        
        Vector<mapped_type> vl()
        {
            return values();
        }

        /**
            Checks whether map contains specific key.
        */
        bool contains(const key_type& key)
        {
            return m_map.find(key) != m_map.end();
        }
        
        bool ct(const key_type& key)
        {
            return contains(key);
        }

        size_type sz()
        {
            return size();
        }

        /**
            Removes an element by key from the map and returns reference to itself.
        */
        Map<Key, Mapped>& operator-(const Key& item)
        {
            erase(item);
            return *this;
        }

    private:
        underlaying_container m_map;

    public:
        detail::getter<size_type> c;
        detail::getter<Vector<mapped_type>> v;

    private:
        void update_getters()
        {
            c = decltype(c)(std::bind(&Map::size, this)); // 'c'ount
            v = decltype(v)(std::bind(&Map::values, this)); // 'v'alues
        }
    };

    //shortened
    template<typename K, typename V>
    using M = Map<K, V>;
}
