﻿#pragma once
#include <string>
#include <functional>
#include "numerics.hpp"
#include "utils/getter.hpp"
#include "utils/circular_indexer.hpp"
#include "vector.hpp"

namespace swerstl
{
    // ToDo: substring and replace
    // ToDo: find

    /**
        Wrapper around std::string (std::basic_string<Character>)
    */
    class String
    {
    public:
#ifdef _MSC_VER
#pragma region CPPSTL
#endif
        using underlaying_container = ::std::basic_string<Character>;
        using value_type = underlaying_container::value_type;
        using traits_type = underlaying_container::traits_type;
        using allocator_type = underlaying_container::allocator_type;
        using reference = underlaying_container::reference;
        using const_reference = underlaying_container::const_reference;
        using pointer = underlaying_container::pointer;
        using const_pointer = underlaying_container::const_pointer;
        using iterator = underlaying_container::iterator;
        using const_iterator = underlaying_container::const_iterator;
        using reverse_iterator = underlaying_container::reverse_iterator;
        using const_reverse_iterator = underlaying_container::const_reverse_iterator;
        using difference_type = underlaying_container::difference_type;
        using size_type = underlaying_container::size_type;

        static constexpr size_type npos = static_cast<size_type>(-1); // std::string::npos sadly isn't constexpr

        String(const String& other) : m_string(other.m_string)
        {
            update_getters();
        }
        String(String&& other) : m_string(std::move(other.m_string))
        {
            update_getters();
        }
        String& operator=(const String& other)
        {
            if (this != &other)
            {
                m_string = other.m_string;
                update_getters();
            }
            return *this;
        }
        String& operator=(String&& other)
        {
            if (this != &other)
            {
                m_string = std::move(other.m_string);
                update_getters();
            }
            return *this;
        }

        ~String() = default;


        String() : m_string()
        {
            update_getters();
        }
        explicit String(const std::string& str) : m_string(str)
        {
            update_getters();
        }
        String(const char* str) : m_string(str)
        {
            update_getters();
        }
        String(std::initializer_list<char> il) : m_string(il)
        {
            update_getters();
        }
        String(size_type n, value_type c) : m_string(n, c)
        {
            update_getters();
        }

        String& operator=(char c)
        {
            m_string = c;
            update_getters();
            return *this;
        }

        iterator begin() { return m_string.begin(); }
        const_iterator begin() const { return m_string.begin(); }
        iterator end() { return m_string.end(); }
        const_iterator end() const { return m_string.end(); }
        reverse_iterator rbegin() noexcept { return m_string.rbegin(); }
        const_reverse_iterator rbegin() const noexcept { return m_string.rbegin(); }
        reverse_iterator rend() { return m_string.rend(); }
        const_reverse_iterator rend() const { return m_string.rend(); }

        size_type size() const noexcept { return m_string.size(); }
        size_type length() const noexcept { return m_string.length(); }
        size_type max_size() const { return m_string.max_size(); }
        void resize(size_type n) { m_string.resize(n); }
        void resize(size_type n, char c) { m_string.resize(n, c); }
        void reserve(size_type n = 0) { m_string.reserve(n); }
        void clear() { m_string.clear(); }
        bool empty() const { return m_string.empty(); }
        void shrink_to_fit() { m_string.shrink_to_fit(); }

        reference at(size_type pos) { return m_string.at(pos); }
        const_reference at(size_type pos) const { return m_string.at(pos); }
        reference back() { return m_string.back(); }
        const_reference back() const { return m_string.back(); }
        reference front() { return m_string.front(); }
        const_reference front() const { return m_string.front(); }

        String& operator+=(value_type c)
        {
            m_string += c;
            return *this;
        }
        String& operator+=(const String& str)
        {
            m_string += str.m_string;
            return *this;
        }

        void push_back(value_type c) { m_string.push_back(c); }
        void pop_back() { m_string.pop_back(); }

        String& insert(size_type pos, size_type n, char c) { m_string.insert(pos, n, c); return *this; }
        String& insert(size_type pos, const String& str) { m_string.insert(pos, str.m_string); return *this; }

        String& erase(size_type pos = 0, size_type len = npos) { m_string.erase(pos, len); return *this; }

        String& replace(size_type pos, size_type len, const String& str) { m_string.replace(pos, len, str.m_string); return *this; }

        void swap(String& str)
        {
            m_string.swap(str.m_string);
            update_getters();
            str.update_getters();
        }

        size_type find(value_type c, size_type pos = 0) const noexcept { return m_string.find(c, pos); }
        size_type find(const String& str, size_type pos = 0) const noexcept { return m_string.find(str.m_string, pos); }

        size_type rfind(value_type c, size_type pos = npos) const noexcept { return m_string.rfind(c, pos); }
        size_type rfind(const String& str, size_type pos = npos) const noexcept { return m_string.rfind(str.m_string, pos); }

        size_type find_first_of(value_type c, size_type pos = 0) const noexcept { return m_string.find_first_of(c, pos); }
        size_type find_first_of(const String& str, size_type pos = 0) const noexcept { return m_string.find_first_of(str.m_string, pos); }

        size_type find_last_of(value_type c, size_type pos = npos) const noexcept { return m_string.find_last_of(c, pos); }
        size_type find_last_of(const String& str, size_type pos = npos) const noexcept { return m_string.find_last_of(str.m_string, pos); }

        size_type find_first_not_of(value_type c, size_type pos = 0) const noexcept { return m_string.find_first_not_of(c, pos); }
        size_type find_first_not_of(const String& str, size_type pos = 0) const noexcept { return m_string.find_first_not_of(str.m_string, pos); }

        size_type find_last_not_of(value_type c, size_type pos = npos) const noexcept { return m_string.find_last_not_of(c, pos); }
        size_type find_last_not_of(const String& str, size_type pos = npos) const noexcept { return m_string.find_last_not_of(str.m_string, pos); }

        String substr(size_type pos = 0, size_type len = npos) const { return String(m_string.substr(pos, len)); }

        int compare(const String& str) const noexcept { return m_string.compare(str.m_string); }
        int compare(size_type pos, size_type len, const String& str) const { return m_string.compare(pos, len, str.m_string); }
        int compare(size_type pos, size_type len, const String& str, size_type subpos, size_type sublen = npos) const { return m_string.compare(pos, len, str.m_string, subpos, sublen); }

        friend String operator+(const String&  lhs, const String&  rhs) { return String(operator+(lhs.m_string, rhs.m_string)); }
        friend String operator+(      String&& lhs,       String&& rhs) { return String(operator+(std::move(lhs.m_string), std::move(rhs.m_string))); }
        friend String operator+(      String&& lhs, const String&  rhs) { return String(operator+(std::move(lhs.m_string), rhs.m_string)); }
        friend String operator+(const String&  lhs,       String&& rhs) { return String(operator+(lhs.m_string, std::move(rhs.m_string))); }

        friend String operator+(const String&  lhs,   value_type   rhs) { return String(operator+(lhs.m_string, rhs)); }
        friend String operator+(      String&& lhs,   value_type   rhs) { return String(operator+(lhs.m_string, rhs)); }
        friend String operator+(  value_type   lhs, const String&  rhs) { return String(operator+(lhs, rhs.m_string)); }
        friend String operator+(  value_type   lhs,       String&& rhs) { return String(operator+(lhs, rhs.m_string)); }

        friend bool operator==(const String& lhs, const String& rhs) noexcept { return operator==(lhs.m_string, rhs.m_string); }
        friend bool operator!=(const String& lhs, const String& rhs) noexcept { return operator!=(lhs.m_string, rhs.m_string); }
        friend bool operator< (const String& lhs, const String& rhs) noexcept { return operator <(lhs.m_string, rhs.m_string); }
        friend bool operator<=(const String& lhs, const String& rhs) noexcept { return operator<=(lhs.m_string, rhs.m_string); }
        friend bool operator> (const String& lhs, const String& rhs) noexcept { return operator >(lhs.m_string, rhs.m_string); }
        friend bool operator>=(const String& lhs, const String& rhs) noexcept { return operator>=(lhs.m_string, rhs.m_string); }

        friend void swap(String& x, String& y) { x.swap(y); }

        friend std::ostream& operator<<(std::ostream& os, const String& str) { return operator<<(os, str.m_string); }
        friend std::istream& operator>>(std::istream& is, String& str) { return operator>>(is, str.m_string); }
        friend std::istream& getline(std::istream& is, String& str) { return std::getline(is, str.m_string); }
        friend std::istream& getline(std::istream&& is, String& str) { return std::getline(std::move(is), str.m_string); }
#ifdef _MSC_VER
#pragma endregion CPPSTL
#endif

        /**
            Constructs a string from given Integer.
        */
        String(Integer number) : String(std::to_string(number))
        {
        }

        /**
            Constructs a string from given Floating.
        */
        String(Floating number) : String(std::to_string(number))
        {
        }

        /**
            Multiplies a string.
            3 * "ABC" == "ABCABCABC"
        */
        friend String operator*(Integer count, const String& str)
        {
            String res = str;
            res *= count;
            return res;
        }
        friend String operator*(const String& str, Integer count) { return operator*(count, str); }

        friend String operator*(Integer count, String&& str)
        {
            String res = std::move(str);
            res *= count;
            return res;
        }
        friend String operator*(String&& str, Integer count) { return operator*(count, std::move(str)); }

        String& operator*=(Integer count)
        {
            String str = *this;
            for (int i = 1; i < count; ++i)
                *this += str;
            return *this;
        }

        /**
            Accesses an element using circular_indexer.
        */
        reference operator[](difference_type pos)
        {
            return detail::circular_indexer(m_string, pos);
        }
        const_reference operator[](difference_type pos) const
        {
            return detail::circular_indexer(m_string, pos);
        }

        /**
            Returns new String, which has its characters in reversed order.
        */
        String reverse() const
        {
            String other = *this;
            std::reverse(other.begin(), other.end());
            return other;
        }

        String rv() const
        {
            return reverse();
        }

        /**
            Splits a String by delimiters (each character may be separator) and
            returns a Vector with separated parts of the String.
        */
        Vector<String> split(String delimiters = ",") const
        {
            size_type start = 0, length;
            Vector<String> result;
            while (npos != (length = find_first_of(delimiters, start)))
            {
                result += substr(start, length - start);
                start = length + 1;
            }
            result += substr(start);
            return result;
        }

        Vector<String> sp(String delimiters = ",") const
        {
            return split(delimiters);
        }

        /**
            Converts a String to Integer.
        */
        Integer to_integer() const
        {
            return std::stoll(m_string);
        }

        /**
            Converts a String to Vector<Character> vec; where vec[i] = operator[](i) for all i in [0, size()]
        */
        Vector<Character> to_vector() const
        {
            Vector<Character> v;
            for (const_reference c : *this)
                v.push_back(c);
            return v;
        }

    private:
        std::string m_string;

    public:
        const underlaying_container& get_underlaying_container() const { return m_string; }
        detail::getter<I> c;
        detail::getter<String> r;
        detail::getter<Vector<String>> s;
        detail::getter<Integer> i;
        detail::getter<Vector<Character>> v;

    private:
        // std::function copy ctor copies bound fnc with pointer to other this instance.
        void update_getters()
        {
            c = decltype(c)(std::bind(&String::size, this)); // 'c'ount
            r = decltype(r)(std::bind(&String::reverse, this)); // 'r'everse
            s = decltype(s)(std::bind(&String::split, this, " ")); // 's'plit
            i = decltype(i)(std::bind(&String::to_integer, this)); // to 'i'nteger
            v = decltype(v)(std::bind(&String::to_vector, this)); // to 'v'ector
        }
    };

    using S = String;
}

/**
    Specializations of standard templates, that may be used in Standard Library code.
*/
namespace std
{
    template<>
    struct hash<swerstl::String>
    {
        size_t operator()(const swerstl::String& x) const
        {
            return std::hash<std::string>()(x.get_underlaying_container());
        }
    };

    template<>
    struct less<swerstl::String>
    {
        bool operator()(const swerstl::String& lhs, const swerstl::String& rhs) const
        {
            return less<std::string>()(lhs.get_underlaying_container(), rhs.get_underlaying_container());
        }
    };
}
