﻿#pragma once
#include <deque>
#include <ostream>
#include <string>
#include <algorithm>
#include <sstream>
#include <iostream>
#include "utils/getter.hpp"
#include "utils/circular_indexer.hpp"
#include "numerics.hpp"

namespace swerstl
{

    // ToDo: find

    /**
        Wrapper around std::deque to provide random access as well as insertion and removal from both ends.
    */
    template<typename T>
    class Vector
    {
    public:
#ifdef _MSC_VER
#pragma region CPPSTL
#endif
        using underlaying_container = std::deque<T>;
        using value_type = typename underlaying_container::value_type;
        using reference = typename underlaying_container::reference;
        using const_reference = typename underlaying_container::const_reference;
        using pointer = typename underlaying_container::pointer;
        using const_pointer = typename underlaying_container::const_pointer;
        using iterator = typename underlaying_container::iterator;
        using const_iterator = typename underlaying_container::const_iterator;
        using reverse_iterator = typename underlaying_container::reverse_iterator;
        using const_reverse_iterator = typename underlaying_container::const_reverse_iterator;
        using difference_type = typename underlaying_container::difference_type;
        using size_type = typename underlaying_container::size_type;
    private:
        explicit Vector(const underlaying_container& x) : m_deque(x)
        {
            update_getters();
        }
        explicit Vector(underlaying_container&& x) : m_deque(std::move(x))
        {
            update_getters();
        }
    public:
        Vector(const Vector<T>& other) : m_deque(other.m_deque)
        {
            update_getters();
        }
        Vector(Vector<T>&& other) : m_deque(std::move(other.m_deque))
        {
            update_getters();
        }
        Vector<T>& operator=(const Vector<T>& other)
        {
            if (this != &other)
            {
                m_deque = other.m_deque;
                update_getters();
            }
            return *this;
        }
        Vector<T>& operator=(Vector<T>&& other)
        {
            if (this != &other)
            {
                m_deque = std::move(other.m_deque);
                update_getters();
            }
            return *this;
        }
        ~Vector() = default;

        Vector() : m_deque()
        {
            update_getters();
        }
        explicit Vector(size_type n) : m_deque(n)
        {
            update_getters();
        }
        Vector(size_type n, const value_type& val) : m_deque(n, val)
        {
            update_getters();
        }
        Vector(std::initializer_list<value_type> il) : m_deque(il)
        {
            update_getters();
        }

        iterator begin() noexcept { return m_deque.begin(); }
        const_iterator begin() const noexcept { return m_deque.begin(); }
        iterator end() noexcept { return m_deque.end(); }
        const_iterator end() const noexcept { return m_deque.end(); }
        reverse_iterator rbegin() noexcept { return m_deque.rbegin(); }
        const_reverse_iterator rbegin() const noexcept { return m_deque.rbegin(); }
        reverse_iterator rend() { return m_deque.rend(); }
        const_reverse_iterator rend() const { return m_deque.rend(); }

        size_type size() const noexcept { return m_deque.size(); }
        size_type max_size() const noexcept { return m_deque.max_size(); }
        void resize(size_type n) { m_deque.resize(n); }
        void resize(size_type n, const value_type& val) { m_deque.resize(n, val); }
        bool empty() const noexcept { return m_deque.empty(); }
        void shrink_to_fit() { m_deque.shrink_to_fit(); }

        reference at(size_type n) { return m_deque.at(n); }
        const_reference at(size_type n) const { return m_deque.at(n); }
        reference front() { return m_deque.front(); }
        const_reference front() const { return m_deque.front(); }
        reference back() { return m_deque.back(); }
        const_reference back() const { return m_deque.back(); }

        void push_back(const value_type& val) { m_deque.push_back(val); }
        void pop_back() { m_deque.pop_back(); }
        void push_front(const value_type& val) { m_deque.push_front(val); }
        void pop_front() { m_deque.pop_front(); }

        iterator insert(const_iterator position, const value_type& val) { return m_deque.insert(position, val); }
        iterator insert(const_iterator position, size_type n, const value_type& val) { return m_deque.insert(position, n, val); }

        iterator erase(const_iterator position) { return m_deque.erase(position); }
        iterator erase(const_iterator first, const_iterator last) { return m_deque.erase(first, last); }

        void clear() noexcept { m_deque.clear(); }
        void swap(Vector& other)
        {
            m_deque.swap(other.m_deque);
            update_getters();
            other.update_getters();
        }

        friend void swap(Vector<T>& lhs, Vector<T>& rhs) { lhs.swap(rhs); }
        friend bool operator==(const Vector<T>& lhs, const Vector<T>& rhs) { return operator==(lhs.m_deque, rhs.m_deque); }
        friend bool operator!=(const Vector<T>& lhs, const Vector<T>& rhs) { return operator!=(lhs.m_deque, rhs.m_deque); }
        friend bool operator< (const Vector<T>& lhs, const Vector<T>& rhs) { return operator< (lhs.m_deque, rhs.m_deque); }
        friend bool operator<=(const Vector<T>& lhs, const Vector<T>& rhs) { return operator<=(lhs.m_deque, rhs.m_deque); }
        friend bool operator> (const Vector<T>& lhs, const Vector<T>& rhs) { return operator> (lhs.m_deque, rhs.m_deque); }
        friend bool operator>=(const Vector<T>& lhs, const Vector<T>& rhs) { return operator>=(lhs.m_deque, rhs.m_deque); }
#ifdef _MSC_VER
#pragma endregion
#endif
    private:
        /**
            Returns const_iterator from given index by advancing from begin (if index >= 0)
            or de-advancing from the end.
        */
        const_iterator iterator_from_index(difference_type index) const
        {
            const_iterator iter = index >= 0 ? begin() : end();
            std::advance(iter, index >= 0 ? index : index + 1);
            return iter;
        }

    public:
        /**
            Combines two Vectors togehter.
        */
        friend Vector<T> operator+(const Vector<T>& lhs, const Vector<T>& rhs)
        {
            Vector<T> result = lhs;
            result += rhs;
            return result;
        }

        /**
            Prepends value to vec and returns it.
        */
        friend Vector<T>& operator+=(const_reference value, Vector<T>& vec)
        {
            vec.push_front(value);
            return vec;
        }

        /**
            Appends value to vector and returns it.
        */
        Vector<T>& operator+=(const_reference value)
        {
            push_back(value);
            return *this;
        }

        /**
            Appends values in order of rhs and returns itself.
        */
        Vector<T>& operator+=(const Vector<T>& rhs) &
        {
            for (const_reference item : rhs)
                push_back(item);
            return *this;
        }

        /**
            Prepends values of lhs in reversed order to rhs, so that they are in order and returns rhs.
        */
        friend Vector<T>& operator+=(const Vector<T>& lhs, Vector<T>& rhs)
        {
            for (auto iter = lhs.rbegin(); iter != lhs.rend(); ++iter)
                rhs.push_front(*iter);
            return rhs;
        }

        /**
            Returns an element at position pos by using circular indexer.
        */
        reference operator[](Integer pos)
        {
            return detail::circular_indexer(m_deque, pos);
        }
        const_reference operator[](Integer pos) const
        {
            return detail::circular_indexer(m_deque, pos);
        }

        /**
            Output stream operator, which outputs "vec[0] vec[1] ... vec[vec.size()-1]"
        */
        friend std::ostream& operator<<(std::ostream& str, const Vector<T>& vec)
        {
            static std::string delimiter = " ";
            if (!vec.empty())
            {
                str << vec.front();
                for (auto iter = std::next(vec.begin()); iter != vec.end(); ++iter)
                    str << delimiter << *iter;
            }
            return str;
        }

        /**
            Reads a vector from stream, supposing format "element1 element2 ... elementN"
        */
        friend std::istream& operator>>(std::istream& stream, Vector<T>& vec)
        {
            vec.clear();
            std::string line;
            std::getline(std::cin, line);
            std::istringstream str(line);
            T item;
            while (str >> item)
                vec.push_back(item);
            return stream;
        }

        /**
            Inserts an item after given index.
        */
        void insert(difference_type index, const_reference item)
        {
            insert(iterator_from_index(index), item);
        }

        /**
            Inserts items at given index in order of items.
        */
        void insert(difference_type index, const Vector<T>& items)
        {
            if (index >= 0)
                for (const_reference item : items)
                    insert(index++, item);
            else
                for (const_reference item : items)
                    insert(index, item);
        }

        /**
            Removes an element at given index.
        */
        void erase(difference_type index)
        {
            erase(iterator_from_index(index));
        }

        /**
            Removes elements from given range [from, to).
        */
        void erase(difference_type from, difference_type to)
        {
            const_iterator ifrom = iterator_from_index(from);
            const_iterator ito = iterator_from_index(to);
            if (ifrom > ito)
                std::swap(ifrom, ito);
            erase(ifrom, ito);
        }

        /**
            Returns new vector, which has elements of *this in sorted order, using std::less.
        */
        Vector<T> sort() const
        {
            Vector<T> other = *this;
            std::sort(other.begin(), other.end());
            return other;
        }
        Vector<T> sr() const
        {
            return sort();
        }

        /**
            Returns new vector, which contains all elements of *this in reversed order.
        */
        Vector<T> reverse() const
        {
            Vector<T> other = *this;
            std::reverse(other.begin(), other.end());
            return other;
        }

        Vector<T> rv() const
        {
            return reverse();
        }

        Integer sz() const
        {
            return size();
        }

        void in(difference_type index, const_reference item)
        {
            return insert(index, item);
        }
        void in(difference_type index, Vector<T> items)
        {
            return insert(index, items);
        }

        void rm(difference_type index)
        {
            return erase(index);
        }
        void rm(difference_type from, difference_type to)
        {
            return erase(from, to);
        }
        Vector<T>& operator-(const T& item)
        {
            erase(std::find(begin(), end(), item));
            return *this;
        }

    private:
        underlaying_container m_deque;

    public:
        const underlaying_container& get_underlaying_container() const { return m_deque; }
        detail::getter<Integer> c;
        detail::getter<Vector<T>> s;
        detail::getter<Vector<T>> r;

    private:
        void update_getters()
        {
            c = decltype(c)(std::bind(&Vector::size, this)); // 'c'ount
            s = decltype(s)(std::bind(&Vector::sort, this)); // 's'ort
            r = decltype(r)(std::bind(&Vector::reverse, this)); // 'r'everse
        }
    };
    
    // shortened
    template<typename T>
    using V = Vector<T>;
}
