﻿#define _SCL_SECURE_NO_WARNINGS
#include <string>
#include <stdexcept>
#include <iostream>
#include <cstring>
#include <jvmti.h>
#include <algorithm>
#include <memory>
#include <vector>
#include <unordered_set>
#include <unordered_map>
#include <fstream>
#include "make_unique.hpp"

#include "bci_server.hpp"
#include "bci_client.hpp"
#include <iterator>
#include <chrono>

void check(jvmtiError err, const std::string& msg)
{
    if (err != JVMTI_ERROR_NONE)
        throw std::runtime_error(msg + "[" + std::to_string(err) + "]");
}

void check(jint err, const std::string& msg)
{
    if (err != JNI_OK)
        throw std::runtime_error(msg + "[" + std::to_string(err) + "]");
}

class stopwatch
{
public:
    void start()
    {
        m_start = std::chrono::steady_clock::now();
    }

    void stop()
    {
        m_end = std::chrono::steady_clock::now();
    }

    template<typename Duration>
    double ellapsed()
    {
        std::chrono::duration<double, typename Duration::period> el = m_end - m_start;
        return el.count();
    }
private:
    std::chrono::time_point<std::chrono::steady_clock> m_start;
    std::chrono::time_point<std::chrono::steady_clock> m_end;
};

stopwatch startup;
stopwatch execution;
std::uint64_t events_count;

void report(std::ostream& stream)
{
    stream << "Startup: " << startup.ellapsed<std::chrono::milliseconds>() << " ms" << std::endl;
    stream << "Execution: " << execution.ellapsed<std::chrono::milliseconds>() << " ms" << std::endl;
    stream << "Events: " << 2 * events_count << std::endl;
}

template<typename F>
static auto check(F f, JNIEnv* jni, const std::string& message) -> typename std::enable_if
<
    std::is_same<decltype(f()), void>::value
>::type
{
    f();
    if (jni->ExceptionCheck())
        throw std::runtime_error(message);
}

template<typename F>
static auto check(F f, JNIEnv* jni, const std::string& message) -> typename std::enable_if
<
    !std::is_same<decltype(f()), void>::value,
    decltype(f())
>::type
{
    auto val = f();
    if (jni->ExceptionCheck())
        throw std::runtime_error(message);
    return val;
}

jvmtiEnv* jvmti;
std::string benchmark_class;
std::string benchmark_package;
std::unique_ptr<runtime_checker::instrumentation_server> bci_server;
bool method_hooks_enabled = false;
std::unordered_set<std::string> instrumented_classes;

std::vector<std::string> blacklist_patterns
{
    "Ljava/lang/Thread",
    "Lsun/reflect/"
};

bool is_blacklisted(const std::string& str)
{
    bool b = blacklist_patterns.end() != std::find_if(blacklist_patterns.begin(), blacklist_patterns.end(), [&str](const std::string& text)
    {
        return str.find(text) != std::string::npos;
    });
    return b;
}

std::string get_class_name(jclass klass)
{
    char* name;
    check(jvmti->GetClassSignature(klass, &name, nullptr), "GetClassSignature");
    std::string str(name);
    check(jvmti->Deallocate(reinterpret_cast<unsigned char*>(name)), "Deallocate");
    return str;
}

std::string get_method_name(jmethodID method)
{
    char* name;
    check(jvmti->GetMethodName(method, &name, nullptr, nullptr), "GetClassSignature");
    std::string str(name);
    check(jvmti->Deallocate(reinterpret_cast<unsigned char*>(name)), "Deallocate");
    return str;
}

jclass get_class_of_method(jmethodID method)
{
    jclass klass;
    check(jvmti->GetMethodDeclaringClass(method, &klass), "GetMethodDeclaringClass");
    return klass;
}

jmethodID get_current_method(jint depth)
{
    jvmtiFrameInfo frame;
    jint count;
    check(jvmti->GetStackTrace(nullptr, depth, 1, &frame, &count), "GetStackTrace");
    return frame.method;
}

void enable_event(jvmtiEvent event)
{
    check(jvmti->SetEventNotificationMode(JVMTI_ENABLE, event, nullptr), "SetEventNotificationModeEnable [" + std::to_string(event) + "]");
}

void disable_event(jvmtiEvent event)
{
    check(jvmti->SetEventNotificationMode(JVMTI_DISABLE, event, nullptr), "SetEventNotificationModeDisable [" + std::to_string(event) + "]");
}

extern "C" JNIEXPORT void JNICALL MethodEntry(JNIEnv* jni, jobject)
{
    try
    {
        jmethodID method = get_current_method(1); // ignoring this native method, thus 1
        (void) method;
        ++events_count;
    }
    catch (const std::runtime_error& e)
    {
        std::cout << e.what() << std::endl;
    }
}

extern "C" JNIEXPORT void JNICALL MethodExit(JNIEnv* jni, jobject, jobject ret_value)
{
    try
    {
        jmethodID method = get_current_method(1); // ignoring this native method, thus 1
        (void) method;
    }
    catch (const std::runtime_error& e)
    {
        std::cout << e.what() << std::endl;
    }
}

void JNICALL ClassPrepare(jvmtiEnv*, JNIEnv* jni, jthread thread, jclass klass)
{
    try
    {
        jint mods;
        check(jvmti->GetClassModifiers(klass, &mods), "GetClassModifiers");
        if (mods & 0x0200) // interface
            return;

        std::string class_name = get_class_name(klass);
        if (!is_blacklisted(class_name) && class_name.find("L" + benchmark_package) != std::string::npos)
        {
            JNINativeMethod ms[] = {
            { "__MethodEntryHook__", "()V", &MethodEntry },
            { "__MethodExitHook__", "(Ljava/lang/Object;)V", &MethodExit }
            };
            check(jni->RegisterNatives(klass, ms, sizeof ms / sizeof ms[0]), "RegisterNatives " + class_name);
        }

        if (!method_hooks_enabled && class_name == "L" + benchmark_class + ";")
        {
            method_hooks_enabled = true;
            startup.stop();
            execution.start();

            jclass objclass = check([=] { return jni->FindClass("java/lang/Object"); }, jni, "FindClass");
            jfieldID field = check([=] { return jni->GetStaticFieldID(objclass, "__JVMTIHooksEnabled__", "Z"); }, jni, "FindClass");
            check([=] { jni->SetStaticBooleanField(objclass, field, true); }, jni, "SetStaticBooleanField");
        }

    }
    catch (const std::runtime_error& e)
    {
        std::cout << e.what() << std::endl;
    }
}

void JNICALL ClassFileLoadHook(jvmtiEnv*, JNIEnv* jni, jclass klass, jobject loader, const char* name, jobject, jint class_data_len, const unsigned char* class_data, jint* new_class_data_len, unsigned char** new_class_data)
{
    try
    {
        if (!is_blacklisted(std::string("L") + name) && (std::string(name).find(benchmark_package) != std::string::npos || std::string(name) == "java/lang/Object"))
        {
            runtime_checker::instrumentation_client bci_client(name, class_data, class_data_len);
            auto result = bci_client.request_instrumentation(*bci_server);
            if (result.size())
            {
                *new_class_data_len = static_cast<jint>(result.size());
                jvmti->Allocate(*new_class_data_len, new_class_data);
                std::copy(result.data(), result.data() + result.size(), *new_class_data);
                instrumented_classes.insert(std::string("L") + name + ";");
            }
        }
    }
    catch (const std::runtime_error& e)
    {
        std::cout << e.what() << std::endl;
    }
}

void JNICALL VMInit(jvmtiEnv*, JNIEnv* jni, jthread thread)
{
    try
    {
        jint length;
        jclass* classes;
        check(jvmti->GetLoadedClasses(&length, &classes), "GetLoadedClasses");

        JNINativeMethod ms[] = {
            { "__MethodEntryHook__", "()V", &MethodEntry },
            { "__MethodExitHook__", "(Ljava/lang/Object;)V", &MethodExit }
        };

        for (jint i = 0; i < length; ++i)
        {
            auto iter = instrumented_classes.find(get_class_name(classes[i]));
            if (iter != instrumented_classes.end() && *iter != "Ljava/lang/Object;" && !is_blacklisted(*iter))
                check(jni->RegisterNatives(classes[i], ms, sizeof ms / sizeof ms[0]), "RegisterNatives " + *iter);
        }
    }
    catch (const std::runtime_error& err)
    {
        std::cout << err.what() << std::endl;
    }
}

void JNICALL VMDeath(jvmtiEnv*, JNIEnv* jni)
{
    try
    {
        execution.stop();
        report(std::cerr);
    }
    catch (const std::runtime_error& err)
    {
        std::cout << err.what() << std::endl;
    }
}

JNIEXPORT jint JNICALL Agent_OnLoad(JavaVM* vm, char* options, void*)
{
    try
    {
        startup.start();
        if (!options)
            throw std::runtime_error("Expected =bechmark.class.name,benchmark.to.instrument to be present as option value.");

        benchmark_class = options;
        std::replace(benchmark_class.begin(), benchmark_class.end(), '.', '/');
        std::size_t pos = benchmark_class.find(',');
        if (pos == std::string::npos)
            throw std::runtime_error("Expected =bechmark.class.name,benchmark.to.instrument to be present as option value.");

        benchmark_package = benchmark_class.substr(pos + 1);
        benchmark_class.erase(pos);

        check(vm->GetEnv(reinterpret_cast<void**>(&jvmti), JVMTI_VERSION_1_0), "GetEnv");
        
        jvmtiCapabilities capabilities;
        std::memset(&capabilities, 0, sizeof(jvmtiCapabilities));
        capabilities.can_generate_all_class_hook_events = capabilities.can_retransform_classes = true;
        check(jvmti->AddCapabilities(&capabilities), "AddCapabilities");

        jvmtiEventCallbacks callbacks;
        callbacks.ClassPrepare = &ClassPrepare;
        callbacks.ClassFileLoadHook = &ClassFileLoadHook;
        callbacks.VMInit = &VMInit;
        callbacks.VMDeath = &VMDeath;
        check(jvmti->SetEventCallbacks(&callbacks, sizeof(jvmtiEventCallbacks)), "SetEventCallbacks");

        enable_event(JVMTI_EVENT_CLASS_FILE_LOAD_HOOK);
        enable_event(JVMTI_EVENT_CLASS_PREPARE);
        enable_event(JVMTI_EVENT_VM_INIT);
        enable_event(JVMTI_EVENT_VM_DEATH);

        runtime_checker::bci_server_configuration config("bci-server.jar", "127.0.0.1", 15415);
        bci_server = nonstd::make_unique<runtime_checker::instrumentation_server>(config);
    }
    catch (const std::runtime_error& err)
    {
        std::cerr << err.what() << std::endl;
        return JNI_ERR;
    }

    return JNI_OK;
}
