﻿#pragma once
#include "../platform.hpp"

#ifdef PLATFORM_UNIX
#include "../network.hpp"

#include <string>
#include <thread>
#include <chrono>

#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <errno.h>

namespace nonstd
{
    inline connection::connection(std::string address, int port, bool delayed_connect/* = false*/)
        : connection_address{}, connection_socket(0)
    {
        connection_address.sin_family = AF_INET;
        connection_address.sin_port = htons(port);
        if (inet_pton(AF_INET, address.c_str(), &connection_address.sin_addr) <= 0)
            throw std::runtime_error("Couldn't convert address via inet_pton. [Error: " + std::to_string(errno) + "].");

        if (!delayed_connect)
            connect();
    }

    inline connection::~connection()
    {
        disconnect();
    }

    inline void connection::connect()
    {
        if (connection_socket != 0)
            return;

        if ((connection_socket = socket(AF_INET, SOCK_STREAM, 0)) < 0)
            throw std::runtime_error("Couldn't create a socket. [Error: " + std::to_string(errno) + "].");

        if (try_connect(connection_socket, reinterpret_cast<sockaddr*>(&connection_address), sizeof connection_address) < 0)
            throw std::runtime_error("Couldn't connect to the server. [Error: " + std::to_string(errno) + "].");

        setsockopt(connection_socket, SOL_SOCKET, SO_RCVTIMEO, &TIMEOUT, sizeof TIMEOUT);
    }

    inline int connection::try_connect(int sockfd, const sockaddr* addr, socklen_t addrlen)
    {
        int ret;
        int sleep_ms = 200;
        for (int waited = 0; waited < TIMEOUT; waited += sleep_ms, sleep_ms = std::min(sleep_ms * 2, TIMEOUT))
        {
            ret = ::connect(sockfd, addr, addrlen);
            if (ret < 0 && errno == ECONNREFUSED)
                std::this_thread::sleep_for(std::chrono::milliseconds(sleep_ms));
            else
                break;
        }
        return ret;
    }

    inline void connection::disconnect()
    {
        if (connection_socket != 0)
        {
            close(connection_socket);
            connection_socket = 0;
        }
    }

    inline void connection::send(const connection::byte* data, std::size_t size) const
    {
        int res = ::write(connection_socket, data, size);
        if (res == -1)
            throw std::runtime_error("Data send failed."); // ToDo: better error
    }

    inline void connection::recv(connection::byte* data, std::size_t size) const
    {
        int remaining = static_cast<int>(size);
        while (remaining)
        {
            int read = ::read(connection_socket, data, remaining);
            if (read == -1)
                throw std::runtime_error("Data recv failed."); // ToDo: better error
            remaining -= read;
            data += read;
        }
    }
} // namespace nonstd
#endif // PLATFORM_WIN
