﻿#pragma once
#include "../platform.hpp"

#ifdef PLATFORM_WIN
#include "../network.hpp"

#include <string>
#include <vector>
#include <stdexcept>

#include <windows.h>
#include <winsock.h>
#include <ws2tcpip.h>

namespace nonstd
{
    inline connection::connection(std::string address, int port, bool delayed_connect/* = false*/)
        : connection_address(nullptr), connection_socket(INVALID_SOCKET)
    {
        try
        {
            if (connection_counter++ == 0)
                static_init();
            
            addrinfo hints;
            std::memset(&hints, 0, sizeof hints);
            hints.ai_family = AF_UNSPEC;
            hints.ai_socktype = SOCK_STREAM;
            hints.ai_protocol = IPPROTO_TCP;

            std::string portstr = std::to_string(port);
            int res = getaddrinfo(address.c_str(), portstr.c_str(), &hints, &connection_address);
            if (res != 0)
                throw std::runtime_error("Can't resolve address."); // ToDo: better error

            if (!delayed_connect)
                connect();
        }
        catch (const std::runtime_error&)
        {
            if (--connection_counter == 0)
                static_destroy();
            throw;
        }
    }

    inline void connection::static_init()
    {
        // Initialize Winsock
        WSADATA wsaData;
        if (WSAStartup(MAKEWORD(2, 2), &wsaData) != 0)
            throw std::runtime_error("Failed to initialize WinSock");
    }

    inline connection::~connection()
    {
        disconnect();

        if (connection_address)
            freeaddrinfo(connection_address);

        if (--connection_counter == 0)
            static_destroy();
    }


    inline void connection::static_destroy()
    {
        WSACleanup();
    }

    inline void connection::connect()
    {
        if (connection_socket != INVALID_SOCKET)
            return;

        // ReSharper disable once CppUnreachableCode
        for (addrinfo* ptr = connection_address; ptr != nullptr; ptr = ptr->ai_next)
        {
            connection_socket = socket(ptr->ai_family, ptr->ai_socktype, ptr->ai_protocol);
            if (connection_socket == INVALID_SOCKET)
                throw std::runtime_error("Invaild socket. Error code: " + std::to_string(WSAGetLastError())); // ToDo: better error

            // Connect to server.
            int res = ::connect(connection_socket, ptr->ai_addr, static_cast<int>(ptr->ai_addrlen));
            if (res != SOCKET_ERROR)
                break;

            closesocket(connection_socket);
            connection_socket = INVALID_SOCKET;
        }

        setsockopt(connection_socket, SOL_SOCKET, SO_RCVTIMEO, reinterpret_cast<const char*>(&TIMEOUT), sizeof TIMEOUT);

        if (connection_socket == INVALID_SOCKET)
            throw std::runtime_error("Can't establish a connection. Error code: " + std::to_string(WSAGetLastError())); // ToDo: better error
    }

    inline void connection::disconnect()
    {
        if (connection_socket != INVALID_SOCKET)
        {
            closesocket(connection_socket);
            connection_socket = INVALID_SOCKET;
        }
    }

    inline void connection::send(const connection::byte* data, std::size_t size) const
    {
        int res = ::send(connection_socket, data, static_cast<int>(size), 0);
        if (res == SOCKET_ERROR)
            throw std::runtime_error("Data send failed [" + std::to_string(WSAGetLastError()) + "]."); // ToDo: better error
    }

    inline void connection::recv(connection::byte* data, std::size_t size) const
    {
        int remaining = static_cast<int>(size);
        while (remaining)
        {
            int read = ::recv(connection_socket, data, remaining, 0);
            if (read < 0)
                throw std::runtime_error("Data recv failed [" + std::to_string(WSAGetLastError()) + "]."); // ToDo: better error
            if (read == 0)
                throw std::runtime_error("Server closed the connection.");
            remaining -= read;
            data += read;
        }
    }
} // namespace nonstd
#endif // PLATFORM_WIN
