#include <string>
#include <stdexcept>
#include <iostream>
#include <cstring>
#include <jvmti.h>
#include <chrono>
#include <algorithm>

void check(jvmtiError err, const std::string& msg)
{
    if (err != JVMTI_ERROR_NONE)
        throw std::runtime_error(msg + "[" + std::to_string(err) + "]");
}

void check(jint err, const std::string& msg)
{
    if (err != JNI_OK)
        throw std::runtime_error(msg + "[" + std::to_string(err) + "]");
}

jvmtiEnv* jvmti;
std::string benchmark_class;

class stopwatch
{
public:
    void start()
    {
        m_start = std::chrono::steady_clock::now();
    }

    void stop()
    {
        m_end = std::chrono::steady_clock::now();
    }

    template<typename Duration>
    double ellapsed()
    {
        std::chrono::duration<double, typename Duration::period> el = m_end - m_start;
        return el.count();
    }
private:
    std::chrono::time_point<std::chrono::steady_clock> m_start;
    std::chrono::time_point<std::chrono::steady_clock> m_end;
};

stopwatch startup;
stopwatch execution;
std::uint64_t events_count;

void report(std::ostream& stream)
{
    stream << "Startup: " << startup.ellapsed<std::chrono::milliseconds>() << " ms" << std::endl;
    stream << "Execution: " << execution.ellapsed<std::chrono::milliseconds>() << " ms" << std::endl;
    stream << "Events: " << 2 * events_count << std::endl;
}

std::string get_class_name(jclass klass)
{
    char* name;
    check(jvmti->GetClassSignature(klass, &name, nullptr), "GetClassSignature");
    std::string str(name);
    check(jvmti->Deallocate(reinterpret_cast<unsigned char*>(name)), "Deallocate");
    return str;
}

std::string get_method_name(jmethodID method)
{
    char* name;
    check(jvmti->GetMethodName(method, &name, nullptr, nullptr), "GetClassSignature");
    std::string str(name);
    check(jvmti->Deallocate(reinterpret_cast<unsigned char*>(name)), "Deallocate");
    return str;
}

jclass get_class_of_method(jmethodID method)
{
    jclass klass;
    check(jvmti->GetMethodDeclaringClass(method, &klass), "GetMethodDeclaringClass");
    return klass;
}

void enable_event(jvmtiEvent event)
{
    check(jvmti->SetEventNotificationMode(JVMTI_ENABLE, event, nullptr), "SetEventNotificationModeEnable [" + std::to_string(event) + "]");
}

void disable_event(jvmtiEvent event)
{
    check(jvmti->SetEventNotificationMode(JVMTI_DISABLE, event, nullptr), "SetEventNotificationModeDisable [" + std::to_string(event) + "]");
}

void JNICALL MethodEntry(jvmtiEnv*, JNIEnv* jni, jthread thread, jmethodID method)
{
    try
    {
        ++events_count;
    }
    catch (const std::runtime_error& err)
    {
        std::cerr << err.what() << std::endl;
    }
}

void JNICALL MethodExit(jvmtiEnv*, JNIEnv* jni, jthread thread, jmethodID method, jboolean by_exception, jvalue ret_value)
{
    try
    {
    }
    catch (const std::runtime_error& err)
    {
        std::cerr << err.what() << std::endl;
    }
}

void JNICALL ClassPrepare(jvmtiEnv*, JNIEnv* jni, jthread thr, jclass klass)
{
    try
    {
        if (get_class_name(klass) == "L" + benchmark_class + ";")
        {
            startup.stop();
            execution.start();
            enable_event(JVMTI_EVENT_METHOD_ENTRY);
            enable_event(JVMTI_EVENT_METHOD_EXIT);
            disable_event(JVMTI_EVENT_CLASS_PREPARE);
        }
    }
    catch (const std::runtime_error& err)
    {
        std::cerr << err.what() << std::endl;
    }
}

void JNICALL VMDeath(jvmtiEnv*, JNIEnv* jni)
{
    try
    {
        execution.stop();
        report(std::cerr);
    }
    catch (const std::runtime_error& err)
    {
        std::cerr << err.what() << std::endl;
    }
}

JNIEXPORT jint JNICALL Agent_OnLoad(JavaVM* vm, char* options, void*)
{
    try
    {
        startup.start();
        if (!options)
            throw std::runtime_error("Expected =bechmark.class.name to be present as option value.");
        benchmark_class = options;
        std::replace(benchmark_class.begin(), benchmark_class.end(), '.', '/');

        check(vm->GetEnv(reinterpret_cast<void**>(&jvmti), JVMTI_VERSION_1_0), "GetEnv");

        jvmtiCapabilities capabilities;
        std::memset(&capabilities, 0, sizeof(jvmtiCapabilities));
        capabilities.can_generate_method_entry_events = capabilities.can_generate_method_exit_events = true;
        check(jvmti->AddCapabilities(&capabilities), "AddCapabilities");

        jvmtiEventCallbacks callbacks;
        callbacks.ClassPrepare = &ClassPrepare;
        callbacks.MethodEntry = &MethodEntry;
        callbacks.MethodExit = &MethodExit;
        callbacks.VMDeath = &VMDeath;
        check(jvmti->SetEventCallbacks(&callbacks, sizeof(jvmtiEventCallbacks)), "SetEventCallbacks");

        enable_event(JVMTI_EVENT_CLASS_PREPARE);
        enable_event(JVMTI_EVENT_VM_DEATH);
    }
    catch (const std::runtime_error& err)
    {
        std::cerr << err.what() << std::endl;
        return JNI_ERR;
    }

    return JNI_OK;
}
